"""
Tests for bimatrix_generators.py

"""
import numpy as np
from scipy.special import comb
from numpy.testing import assert_array_equal, assert_, assert_raises
from quantecon import num_compositions
from quantecon.game_theory import pure_nash_brute

from quantecon.game_theory import (
    blotto_game, ranking_game, sgc_game, tournament_game, unit_vector_game,
)


class TestBlottoGame:
    def setup_method(self):
        self.h, self.t = 4, 3
        rho = 0.5
        self.g = blotto_game(self.h, self.t, rho)

    def test_size(self):
        n = num_compositions(self.h, self.t)
        assert_(self.g.nums_actions == (n, n))

    def test_constant_diagonal(self):
        for i in range(self.g.N):
            diag = self.g.payoff_arrays[i].diagonal()
            assert_((diag == diag[0]).all())

    def test_seed(self):
        seed = 0
        h, t = 3, 4
        rho = -0.5
        for gen in [lambda x: x, np.random.default_rng]:
            gs = [
                blotto_game(h, t, rho, random_state=gen(seed))
                for i in range(2)
            ]
            assert_array_equal(*[g.payoff_profile_array for g in gs])


class TestRankingGame:
    def setup_method(self):
        self.n = 100
        self.g = ranking_game(self.n)

    def test_size(self):
        assert_(self.g.nums_actions == (self.n, self.n))

    def test_weakly_decreasing_row_wise_payoffs(self):
        for payoff_array in self.g.payoff_arrays:
            assert_((payoff_array[:, 1:-1] >= payoff_array[:, 2:]).all())

    def test_elements_first_row(self):
        assert_(
            self.g.payoff_arrays[0][0, 0] + self.g.payoff_arrays[1][0, 0] == 1.
        )
        for payoff_array in self.g.payoff_arrays:
            assert_(np.isin(payoff_array[0, :], [0, 1, 0.5]).all())

    def test_seed(self):
        seed = 0
        n = 100
        for gen in [lambda x: x, np.random.default_rng]:
            gs = [ranking_game(n, random_state=gen(seed)) for i in range(2)]
            assert_array_equal(*[g.payoff_profile_array for g in gs])


def test_sgc_game():
    k = 2
    s = """\
        0.750 0.750 1.000 0.500 0.500 1.000 0.000 0.500 0.000 0.500 0.000 0.500
        0.000 0.500 0.500 1.000 0.750 0.750 1.000 0.500 0.000 0.500 0.000 0.500
        0.000 0.500 0.000 0.500 1.000 0.500 0.500 1.000 0.750 0.750 0.000 0.500
        0.000 0.500 0.000 0.500 0.000 0.500 0.500 0.000 0.500 0.000 0.500 0.000
        0.750 0.000 0.000 0.750 0.000 0.000 0.000 0.000 0.500 0.000 0.500 0.000
        0.500 0.000 0.000 0.750 0.750 0.000 0.000 0.000 0.000 0.000 0.500 0.000
        0.500 0.000 0.500 0.000 0.000 0.000 0.000 0.000 0.750 0.000 0.000 0.750
        0.500 0.000 0.500 0.000 0.500 0.000 0.000 0.000 0.000 0.000 0.000 0.750
        0.750 0.000"""
    bimatrix = np.fromstring(s, sep=' ')
    bimatrix.shape = (4*k-1, 4*k-1, 2)
    bimatrix = bimatrix.swapaxes(0, 1)

    g = sgc_game(k)
    assert_array_equal(g.payoff_profile_array, bimatrix)


class TestTournamentGame:
    def setup_method(self):
        self.n = 5
        self.k = 3
        self.m = comb(self.n, self.k, exact=True)
        self.g = tournament_game(self.n, self.k)

    def test_size(self):
        assert_(self.g.nums_actions == (self.n, self.m))

    def test_payoff_values(self):
        possible_values = [0, 1]
        for payoff_array in self.g.payoff_arrays:
            assert_(np.isin(payoff_array, possible_values).all())

        max_num_dominated_subsets = \
            sum([comb(i, self.k, exact=True) for i in range(self.n)])
        assert_(self.g.payoff_arrays[0].sum() <= max_num_dominated_subsets)
        assert_((self.g.payoff_arrays[1].sum(axis=1) == self.k).all())

    def test_seed(self):
        seed = 0
        for gen in [lambda x: x, np.random.default_rng]:
            gs = [
                tournament_game(self.n, self.k, random_state=gen(seed))
                for i in range(2)
            ]
            assert_array_equal(*[g.payoff_profile_array for g in gs])

    def test_raises_value_error_too_large_inputs(self):
        n, k = 100, 50
        assert_raises(ValueError, tournament_game, n, k)


class TestUnitVectorGame:
    def setup_method(self):
        self.n = 100
        self.g = unit_vector_game(self.n)

    def test_size(self):
        assert_(self.g.nums_actions == (self.n, self.n))

    def test_payoff_values(self):
        # Player 0
        assert_((np.sum(self.g.players[0].payoff_array, axis=0) == 1).all())

    def test_avoid_pure_nash(self):
        NEs = pure_nash_brute(unit_vector_game(self.n, avoid_pure_nash=True),
                              tol=0)
        assert_(len(NEs) == 0)

    def test_seed(self):
        seed = 0
        n = 100
        for gen in [lambda x: x, np.random.default_rng]:
            gs = [
                unit_vector_game(n, random_state=gen(seed)) for i in range(2)
            ]
            assert_array_equal(*[g.payoff_profile_array for g in gs])

    def test_redraw(self):
        seed = 1
        n = 2
        g = unit_vector_game(n, avoid_pure_nash=True, random_state=seed)
        NEs = pure_nash_brute(g, tol=0)
        assert_(len(NEs) == 0)

    def test_raises_value_error_avoid_pure_nash_n_1(self):
        n = 1
        assert_raises(ValueError, unit_vector_game, n, avoid_pure_nash=True)


def test_payoff_range():
    func_kwargs_tuples = \
        [(sgc_game, {'k': 3}),
         (tournament_game, {'n': 6, 'k': 3}),
         (unit_vector_game, {'n': 5}),
         (unit_vector_game, {'n': 5, 'avoid_pure_nash': True})]
    for func, kwargs in func_kwargs_tuples:
        g = func(**kwargs)
        for payoff_array in g.payoff_arrays:
            assert_array_equal(np.clip(payoff_array, 0, 1), payoff_array)



"""
Tests for markov/estimate.py

"""
import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
from quantecon import estimate_mc
from quantecon.markov import fit_discrete_mc


class TestEstimateMCDiscrete:
    def setup_method(self):
        self.test_index_series = np.array((0, 1, 1, 1, 1, 0, 2, 1))
        self.initial_state_values = np.array(['a', 'b', 'c', 'd'])
        self.test_value_series = \
            self.initial_state_values[self.test_index_series]

        self.P = np.array([[0.,   0.5,  0.5 ],
                           [0.25, 0.75, 0.  ],
                           [0.,   1.,   0.  ]])

        self.final_state_indices = np.array([0, 1, 2])
        self.final_state_values = \
            self.initial_state_values[self.final_state_indices]

    def test_integer_state(self):
        mc = estimate_mc(self.test_index_series)
        assert_allclose(mc.P, self.P)
        assert_array_equal(mc.state_values, self.final_state_indices)

    def test_non_integer_state(self):
        mc = estimate_mc(self.test_value_series)
        assert_allclose(mc.P, self.P)
        assert_array_equal(mc.state_values, self.final_state_values)

        mc = estimate_mc(self.test_index_series)
        mc.state_values = self.initial_state_values[mc.state_values]
        assert_allclose(mc.P, self.P)
        assert_array_equal(mc.state_values, self.final_state_values)

    def test_mult_dim_state(self):
        initial_state_values = np.array([[0.97097089, 0.76167618],
                                         [0.61878456, 0.41691572],
                                         [0.42137226, 0.09307409],
                                         [0.62696609, 0.40898893]])
        X = initial_state_values[self.test_index_series]
        ind = np.lexsort(
            np.rot90(initial_state_values[self.final_state_indices])
        )
        final_state_values = initial_state_values[ind]

        mc = estimate_mc(X)
        assert_allclose(mc.P, self.P[np.ix_(ind, ind)])
        assert_array_equal(mc.state_values, final_state_values)


class TestFitDiscreteMC:
    def setup_method(self):
        self.grids = ((np.arange(4), np.arange(5)))
        self.X = [(-0.1, 1.2), (2, 0), (2, 3),
                  (4.4, 4.0), (0.6, 0.4), (1.0, 0.1)]

    def test_order_f(self):
        mc = fit_discrete_mc(self.X, self.grids, order='F')
        S_expected = np.array([[1, 0], [2, 0], [0, 1], [2, 3], [3, 4]])
        P_expected = np.array([[1., 0., 0., 0., 0.],
                               [0., 0., 0., 1., 0.],
                               [0., 1., 0., 0., 0.],
                               [0., 0., 0., 0., 1.],
                               [1., 0., 0., 0., 0.]])

        assert_array_equal(mc.state_values, S_expected)
        assert_allclose(mc.P, P_expected)

    def test_order_c(self):
        mc = fit_discrete_mc(self.X, self.grids, order='C')
        S_expected = np.array([[0, 1], [1, 0], [2, 0], [2, 3], [3, 4]])
        P_expected = np.array([[0., 0., 1., 0., 0.],
                               [0., 1., 0., 0., 0.],
                               [0., 0., 0., 1., 0.],
                               [0., 0., 0., 0., 1.],
                               [0., 1., 0., 0., 0.]])

        assert_array_equal(mc.state_values, S_expected)
        assert_allclose(mc.P, P_expected)



import numpy as np
from numba import njit
from .core import MarkovChain
from .._gridtools import cartesian_nearest_index, cartesian


def estimate_mc(X):
    r"""
    Estimate the Markov chain associated with a time series :math:`X =
    (X_0, \ldots, X_{T-1})` assuming that the state space is the finite
    set :math:`\{X_0, \ldots, X_{T-1}\}` (duplicates removed). The
    estimation is by maximum likelihood. The estimated transition
    probabilities are given by the matrix :math:`P` such that
    :math:`P[i, j] = N_{ij} / N_i`, where :math:`N_{ij} =
    \sum_{t=0}^{T-1} 1_{\{X_t=s_i, X_{t+1}=s_j\}}`, the number of
    transitions from state :math:`s_i` to state :math:`s_j`, while
    :math:`N_i` is the total number of visits to :math:`s_i`. The result
    is returned as a `MarkovChain` instance.

    Parameters
    ----------
    X : array_like
        A time series of state values, from which the transition matrix
        will be estimated, where `X[t]` contains the t-th observation.

    Returns
    -------
    mc : MarkovChain
        A MarkovChain instance where `mc.P` is a stochastic matrix
        estimated from the data `X` and `mc.state_values` is an array of
        values that appear in `X` (sorted in ascending order).

    """
    X = np.asarray(X)
    axis = 0 if X.ndim > 1 else None
    state_values, indices = np.unique(X, return_inverse=True, axis=axis)

    n = len(state_values)
    P = np.zeros((n, n))  # dtype=float to modify in place upon normalization
    P = _count_transition_frequencies(indices, P)
    P /= P.sum(1)[:, np.newaxis]

    mc = MarkovChain(P, state_values=state_values)
    return mc


@njit(cache=True)
def _count_transition_frequencies(index_series, trans_counter):
    T = len(index_series)
    i = index_series[0]
    for t in range(1, T):
        j = index_series[t]
        trans_counter[i, j] += 1
        i = j
    return trans_counter


def fit_discrete_mc(X, grids, order='C'):
    r"""
    Function that takes an arbitrary time series :math: `(X_t)_{t=0}^{T-1}` in
    :math: `\mathbb R^n` plus a set of grid points in each dimension and converts
    it to a MarkovChain by first applying discretization onto the grid
    and then estimation of the Markov chain.

    Parameters
    ----------

    X: array_like(ndim=2)
        Time-series such that the t-th row is :math:`x_t`.
        It should be of the shape T x n, where n is the number of dimensions.

    grids: array_like(array_like(ndim=1))
        Array of `n` sorted arrays. Set of grid points in each dimension

    Examples
    --------

    >>> grids = (np.arange(3), np.arange(2))
    >>> X = [(-0.1, 1.2), (2, 0), (0.6, 0.4), (1.0, 0.1)]
    >>> mc = fit_discrete_mc(X, grids)
    >>> mc.state_values
    array([[0, 1],
           [1, 0],
           [2, 0]])
    >>> mc.P
    array([[0., 0., 1.],
           [0., 1., 0.],
           [0., 1., 0.]])

    Returns
    -------

    mc: MarkovChain
        An instance of the MarkovChain class constructed after discretization
        onto the grid.
    """
    X_indices = cartesian_nearest_index(X, grids, order=order)
    mc = estimate_mc(X_indices)
    # Assign the visited states in the cartesian product as the state values
    prod = cartesian(grids, order=order)
    mc.state_values = prod[mc.state_values]
    return mc



"""
Implements cartesian products and regular cartesian grids, and provides
a function that constructs a grid for a simplex as well as one that
determines the index of a point in the simplex.

"""
import numpy as np
import scipy.special
from numba import jit, njit
from .util.numba import comb_jit


def cartesian(nodes, order='C'):
    '''
    Cartesian product of a list of arrays

    Parameters
    ----------
    nodes : list(array_like(ndim=1))

    order : str, optional(default='C')
        ('C' or 'F') order in which the product is enumerated

    Returns
    -------
    out : ndarray(ndim=2)
        each line corresponds to one point of the product space
    '''

    nodes = [np.asarray(e) for e in nodes]
    shapes = [e.shape[0] for e in nodes]

    dtype = np.result_type(*nodes)

    n = len(nodes)
    l = np.prod(shapes)
    out = np.zeros((l, n), dtype=dtype)

    if order == 'C':
        repetitions = np.cumprod([1] + shapes[:-1])
    else:
        shapes.reverse()
        sh = [1] + shapes[:-1]
        repetitions = np.cumprod(sh)
        repetitions = repetitions.tolist()
        repetitions.reverse()

    for i in range(n):
        _repeat_1d(nodes[i], repetitions[i], out[:, i])

    return out


def mlinspace(a, b, nums, order='C'):
    '''
    Constructs a regular cartesian grid

    Parameters
    ----------
    a : array_like(ndim=1)
        lower bounds in each dimension

    b : array_like(ndim=1)
        upper bounds in each dimension

    nums : array_like(ndim=1)
        number of nodes along each dimension

    order : str, optional(default='C')
        ('C' or 'F') order in which the product is enumerated

    Returns
    -------
    out : ndarray(ndim=2)
        each line corresponds to one point of the product space
    '''

    a = np.asarray(a, dtype='float64')
    b = np.asarray(b, dtype='float64')
    nums = np.asarray(nums, dtype='int64')
    nodes = [np.linspace(a[i], b[i], nums[i]) for i in range(len(nums))]

    return cartesian(nodes, order=order)


@njit
def _repeat_1d(x, K, out):
    '''
    Repeats each element of a vector many times and repeats the whole
    result many times

    Parameters
    ----------
    x : ndarray(ndim=1)
        vector to be repeated

    K : scalar(int)
        number of times each element of x is repeated (inner iterations)

    out : ndarray(ndim=1)
        placeholder for the result

    Returns
    -------
    None
    '''

    N = x.shape[0]
    L = out.shape[0] // (K*N)  # number of outer iterations
    # K                        # number of inner iterations

    # the result out should enumerate in C-order the elements
    # of a 3-dimensional array T of dimensions (K,N,L)
    # such that for all k,n,l, we have T[k,n,l] == x[n]

    for n in range(N):
        val = x[n]
        for k in range(K):
            for l in range(L):
                ind = k*N*L + n*L + l
                out[ind] = val


def cartesian_nearest_index(x, nodes, order='C'):
    """
    Return the index of the point closest to `x` within the cartesian
    product generated by `nodes`. Each array in `nodes` must be sorted
    in ascending order.

    Parameters
    ----------
    x : array_like(ndim=1 or 2)
        Point(s) to search the closest point(s) for.

    nodes : array_like(array_like(ndim=1))
        Array of sorted arrays.

    order : str, optional(default='C')
        ('C' or 'F') order in which the product is enumerated.

    Returns
    -------
    scalar(int) or ndarray(int, ndim=1)
        Index (indices) of the closest point(s) to `x`.

    Examples
    --------
    >>> nodes = (np.arange(3), np.arange(2))
    >>> prod = qe.cartesian(nodes)
    >>> print(prod)
    [[0 0]
     [0 1]
     [1 0]
     [1 1]
     [2 0]
     [2 1]]

    Among the 6 points in the cartesian product `prod`, the closest to
    the point (0.6, 0.4) is `prod[2]`:

    >>> x = (0.6, 0.4)
    >>> qe.cartesian_nearest_index(x, nodes)  # Pass `nodes`, not `prod`
    2

    The closest to (-0.1, 1.2) and (2, 0) are `prod[1]` and `prod[4]`,
    respectively:

    >>> x = [(-0.1, 1.2), (2, 0)]
    >>> qe.cartesian_nearest_index(x, nodes)
    array([1, 4])

    Internally, the index in each dimension is searched by binary search
    and then the index in the cartesian product is calculated (*not* by
    constructing the cartesian product and then searching linearly over
    it).

    """
    x = np.asarray(x)
    is_1d = False
    shape = x.shape
    if len(shape) == 1:
        is_1d = True
        x = x[np.newaxis]
    types = [type(e[0]) for e in nodes]
    dtype = np.result_type(*types)
    nodes = tuple(np.asarray(e, dtype=dtype) for e in nodes)

    n = shape[1-is_1d]
    if len(nodes) != n:
        msg = 'point `x`' if is_1d else 'points in `x`'
        msg += ' must have same length as `nodes`'
        raise ValueError(msg)

    out = _cartesian_nearest_indices(x, nodes, order=order)
    if is_1d:
        return out[0]
    return out


@njit(cache=True)
def _cartesian_nearest_indices(X, nodes, order='C'):
    """
    The main body of `cartesian_nearest_index`, jit-complied by Numba.
    Note that `X` must be a 2-dim ndarray, and a Python list is not
    accepted for `nodes`.

    Parameters
    ----------
    X : ndarray(ndim=2)
        Points to search the closest points for.

    nodes : tuple(ndarray(ndim=1))
        Tuple of sorted ndarrays of same dtype.

    order : str, optional(default='C')
        ('C' or 'F') order in which the product is enumerated.

    Returns
    -------
    ndarray(int, ndim=1)
        Indices of the closest points to the points in `X`.

    """
    m, n = X.shape  # m vectors of length n
    nums_grids = np.empty(n, dtype=np.intp)
    for i in range(n):
        nums_grids[i] = len(nodes[i])

    ind = np.empty(n, dtype=np.intp)
    out = np.empty(m, dtype=np.intp)

    step = -1 if order == 'F' else 1
    slice_ = slice(None, None, step)

    for t in range(m):
        for i in range(n):
            if X[t, i] <= nodes[i][0]:
                ind[i] = 0
            elif X[t, i] >= nodes[i][-1]:
                ind[i] = nums_grids[i] - 1
            else:
                k = np.searchsorted(nodes[i], X[t, i])
                ind[i] = (
                    k if nodes[i][k] - X[t, i] < X[t, i] - nodes[i][k-1]
                    else k - 1
                )
        out[t] = _cartesian_index(ind[slice_], nums_grids[slice_])

    return out


@njit(cache=True)
def _cartesian_index(indices, nums_grids):
    n = len(indices)
    idx = 0
    de_cumprod = 1
    for i in range(1,n+1):
        idx += de_cumprod * indices[n-i]
        de_cumprod *= nums_grids[n-i]
    return idx


_msg_max_size_exceeded = 'Maximum allowed size exceeded'


@jit(nopython=True, cache=True)
def simplex_grid(m, n):
    r"""
    Construct an array consisting of the integer points in the
    (m-1)-dimensional simplex :math:`\{x \mid x_0 + \cdots + x_{m-1} = n
    \}`, or equivalently, the m-part compositions of n, which are listed
    in lexicographic order. The total number of the points (hence the
    length of the output array) is L = (n+m-1)!/(n!*(m-1)!) (i.e.,
    (n+m-1) choose (m-1)).

    Parameters
    ----------
    m : scalar(int)
        Dimension of each point. Must be a positive integer.

    n : scalar(int)
        Number which the coordinates of each point sum to. Must be a
        nonnegative integer.

    Returns
    -------
    out : ndarray(int, ndim=2)
        Array of shape (L, m) containing the integer points in the
        simplex, aligned in lexicographic order.

    Notes
    -----
    A grid of the (m-1)-dimensional *unit* simplex with n subdivisions
    along each dimension can be obtained by `simplex_grid(m, n) / n`.

    Examples
    --------
    >>> simplex_grid(3, 4)
    array([[0, 0, 4],
           [0, 1, 3],
           [0, 2, 2],
           [0, 3, 1],
           [0, 4, 0],
           [1, 0, 3],
           [1, 1, 2],
           [1, 2, 1],
           [1, 3, 0],
           [2, 0, 2],
           [2, 1, 1],
           [2, 2, 0],
           [3, 0, 1],
           [3, 1, 0],
           [4, 0, 0]])

    >>> simplex_grid(3, 4) / 4
    array([[ 0.  ,  0.  ,  1.  ],
           [ 0.  ,  0.25,  0.75],
           [ 0.  ,  0.5 ,  0.5 ],
           [ 0.  ,  0.75,  0.25],
           [ 0.  ,  1.  ,  0.  ],
           [ 0.25,  0.  ,  0.75],
           [ 0.25,  0.25,  0.5 ],
           [ 0.25,  0.5 ,  0.25],
           [ 0.25,  0.75,  0.  ],
           [ 0.5 ,  0.  ,  0.5 ],
           [ 0.5 ,  0.25,  0.25],
           [ 0.5 ,  0.5 ,  0.  ],
           [ 0.75,  0.  ,  0.25],
           [ 0.75,  0.25,  0.  ],
           [ 1.  ,  0.  ,  0.  ]])

    References
    ----------
    A. Nijenhuis and H. S. Wilf, Combinatorial Algorithms, Chapter 5,
    Academic Press, 1978.

    """
    L = num_compositions_jit(m, n)
    if L == 0:  # Overflow occured
    	raise ValueError(_msg_max_size_exceeded)
    out = np.empty((L, m), dtype=np.int_)

    x = np.zeros(m, dtype=np.int_)
    x[m-1] = n

    for j in range(m):
        out[0, j] = x[j]

    h = m

    for i in range(1, L):
        h -= 1

        val = x[h]
        x[h] = 0
        x[m-1] = val - 1
        x[h-1] += 1

        for j in range(m):
            out[i, j] = x[j]

        if val != 1:
            h = m

    return out


def simplex_index(x, m, n):
    r"""
    Return the index of the point x in the lexicographic order of the
    integer points of the (m-1)-dimensional simplex :math:`\{x \mid x_0
    + \cdots + x_{m-1} = n\}`.

    Parameters
    ----------
    x : array_like(int, ndim=1)
        Integer point in the simplex, i.e., an array of m nonnegative
        itegers that sum to n.

    m : scalar(int)
        Dimension of each point. Must be a positive integer.

    n : scalar(int)
        Number which the coordinates of each point sum to. Must be a
        nonnegative integer.

    Returns
    -------
    idx : scalar(int)
        Index of x.

    """
    if m == 1:
        return 0

    decumsum = np.cumsum(x[-1:0:-1])[::-1]
    idx = num_compositions(m, n) - 1
    for i in range(m-1):
        if decumsum[i] == 0:
            break
        idx -= num_compositions(m-i, decumsum[i]-1)
    return idx


def num_compositions(m, n):
    """
    The total number of m-part compositions of n, which is equal to
    (n+m-1) choose (m-1).

    Parameters
    ----------
    m : scalar(int)
        Number of parts of composition.

    n : scalar(int)
        Integer to decompose.

    Returns
    -------
    scalar(int)
        Total number of m-part compositions of n.

    """
    # docs.scipy.org/doc/scipy/reference/generated/scipy.special.comb.html
    return scipy.special.comb(n+m-1, m-1, exact=True)


@jit(nopython=True, cache=True)
def num_compositions_jit(m, n):
    """
    Numba jit version of `num_compositions`. Return `0` if the outcome
    exceeds the maximum value of `np.intp`.

    """
    return comb_jit(n+m-1, m-1)



"""
Utility routines for the markov submodule

"""
import numpy as np
from numba import jit


@jit(nopython=True, cache=True)
def sa_indices(num_states, num_actions):
    """
    Generate `s_indices` and `a_indices` for `DiscreteDP`, for the case
    where all the actions are feasible at every state.

    Parameters
    ----------
    num_states : scalar(int)
        Number of states.

    num_actions : scalar(int)
        Number of actions.

    Returns
    -------
    s_indices : ndarray(int, ndim=1)
        Array containing the state indices.

    a_indices : ndarray(int, ndim=1)
        Array containing the action indices.

    Examples
    --------
    >>> s_indices, a_indices = qe.markov.sa_indices(4, 3)
    >>> s_indices
    array([0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3])
    >>> a_indices
    array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2])

    """
    L = num_states * num_actions
    dtype = np.int_
    s_indices = np.empty(L, dtype=dtype)
    a_indices = np.empty(L, dtype=dtype)

    i = 0
    for s in range(num_states):
        for a in range(num_actions):
            s_indices[i] = s
            a_indices[i] = a
            i += 1

    return s_indices, a_indices


@jit(nopython=True, cache=True)
def _fill_dense_Q(s_indices, a_indices, Q_in, Q_out):
    L = Q_in.shape[0]
    for i in range(L):
        Q_out[s_indices[i], a_indices[i], :] = Q_in[i, :]

    return Q_out


@jit(nopython=True, cache=True)
def _s_wise_max_argmax(a_indices, a_indptr, vals, out_max, out_argmax):
    n = len(out_max)
    for i in range(n):
        if a_indptr[i] != a_indptr[i+1]:
            m = a_indptr[i]
            for j in range(a_indptr[i]+1, a_indptr[i+1]):
                if vals[j] > vals[m]:
                    m = j
            out_max[i] = vals[m]
            out_argmax[i] = a_indices[m]


@jit(nopython=True, cache=True)
def _s_wise_max(a_indices, a_indptr, vals, out_max):
    n = len(out_max)
    for i in range(n):
        if a_indptr[i] != a_indptr[i+1]:
            m = a_indptr[i]
            for j in range(a_indptr[i]+1, a_indptr[i+1]):
                if vals[j] > vals[m]:
                    m = j
            out_max[i] = vals[m]


@jit(nopython=True, cache=True)
def _find_indices(a_indices, a_indptr, sigma, out):
    n = len(sigma)
    for i in range(n):
        for j in range(a_indptr[i], a_indptr[i+1]):
            if sigma[i] == a_indices[j]:
                out[i] = j


@jit(nopython=True, cache=True)
def _has_sorted_sa_indices(s_indices, a_indices):
    """
    Check whether `s_indices` and `a_indices` are sorted in
    lexicographic order.

    Parameters
    ----------
    s_indices, a_indices : ndarray(ndim=1)

    Returns
    -------
    bool
        Whether `s_indices` and `a_indices` are sorted.

    """
    L = len(s_indices)
    for i in range(L-1):
        if s_indices[i] > s_indices[i+1]:
            return False
        if s_indices[i] == s_indices[i+1]:
            if a_indices[i] >= a_indices[i+1]:
                return False
    return True


@jit(nopython=True, cache=True)
def _generate_a_indptr(num_states, s_indices, out):
    """
    Generate `a_indptr`; stored in `out`. `s_indices` is assumed to be
    in sorted order.

    Parameters
    ----------
    num_states : scalar(int)

    s_indices : ndarray(int, ndim=1)

    out : ndarray(int, ndim=1)
        Length must be num_states+1.

    """
    idx = 0
    out[0] = 0
    for s in range(num_states-1):
        while(s_indices[idx] == s):
            idx += 1
        out[s+1] = idx
    out[num_states] = len(s_indices)



"""
Tools for repeated game.

"""

import numpy as np
from scipy.spatial import ConvexHull
from numba import njit


class RepeatedGame:
    """
    Class representing an N-player repeated game.

    Parameters
    ----------
    stage_game : NormalFormGame
        The stage game used to create the repeated game.

    delta : scalar(float)
        The common discount rate at which all players discount the future.

    Attributes
    ----------
    sg : NormalFormGame
        The stage game. See Parameters.

    delta : scalar(float)
        See Parameters.

    N : scalar(int)
        The number of players.

    nums_actions : tuple(int)
        Tuple of the numbers of actions, one for each player.
    """
    def __init__(self, stage_game, delta):
        self.sg = stage_game
        self.delta = delta
        self.N = stage_game.N
        self.nums_actions = stage_game.nums_actions

    def equilibrium_payoffs(self, method=None, options=None):
        """
        Compute the set of payoff pairs of all pure-strategy subgame-perfect
        equilibria with public randomization for any repeated two-player games
        with perfect monitoring and discounting.

        Parameters
        ----------
        method : str, optional
            The method for solving the equilibrium payoff set.

        options : dict, optional
            A dictionary of method options. For example, 'abreu_sannikov'
            method accepts the following options:

                tol : scalar(float)
                    Tolerance for convergence checking.
                max_iter : scalar(int)
                    Maximum number of iterations.
                u_init : ndarray(float, ndim=1)
                    The initial guess of threat points.

        Notes
        -----
        Here lists all the implemented methods. The default method
        is 'abreu_sannikov'.

            1. 'abreu_sannikov'
        """
        if method is None:
            method = 'abreu_sannikov'

        if options is None:
            options = {}

        if method in ('abreu_sannikov', 'AS'):
            return _equilibrium_payoffs_abreu_sannikov(self, **options)
        else:
            msg = f"method {method} not supported."
            raise NotImplementedError(msg)


def _equilibrium_payoffs_abreu_sannikov(rpg, tol=1e-12, max_iter=500,
                                        u_init=np.zeros(2)):
    """
    Using 'abreu_sannikov' algorithm to compute the set of payoff pairs
    of all pure-strategy subgame-perfect equilibria with public randomization
    for any repeated two-player games with perfect monitoring and
    discounting, following Abreu and Sannikov (2014).

    Parameters
    ----------
    rpg : RepeatedGame
        Two player repeated game.

    tol : scalar(float), optional(default=1e-12)
        Tolerance for convergence checking.

    max_iter : scalar(int), optional(default=500)
        Maximum number of iterations.

    u_init : ndarray(float, ndim=1), optional(default=np.zeros(2))
        The initial guess of threat points.

    Returns
    -------
    hull : scipy.spatial.ConvexHull
        The convex hull of equilibrium payoff pairs.

    References
    ----------
    .. [1] Abreu, Dilip, and Yuliy Sannikov. "An algorithm for
       two‐player repeated games with perfect monitoring." Theoretical
       Economics 9.2 (2014): 313-338.
    """
    sg, delta = rpg.sg, rpg.delta

    if sg.N != 2:
        msg = "this algorithm only applies to repeated two-player games."
        raise NotImplementedError(msg)

    best_dev_gains = _best_dev_gains(rpg)
    IC = np.empty(2)
    action_profile_payoff = np.empty(2)
    # auxiliary array for checking if payoff is inside the convex hull
    # first two entries for payoff point, and the last entry is 1.
    extended_payoff = np.ones(3)
    # array to store new points of C in each intersection
    # at most 4 new points will be generated
    new_pts = np.empty((4, 2))
    # array to store the points of W
    # the length of v is limited by |A1|*|A2|*4
    W_new = np.empty((np.prod(sg.nums_actions)*4, 2))
    W_old = np.empty((np.prod(sg.nums_actions)*4, 2))
    # count the new points generated in each iteration
    n_new_pt = 0

    # copy the threat points
    u = np.copy(u_init)

    # initialization
    payoff_pts = \
        sg.payoff_profile_array.reshape(np.prod(sg.nums_actions), 2)
    W_new[:np.prod(sg.nums_actions)] = payoff_pts
    n_new_pt = np.prod(sg.nums_actions)

    n_iter = 0
    while True:
        W_old[:n_new_pt] = W_new[:n_new_pt]
        n_old_pt = n_new_pt
        hull = ConvexHull(W_old[:n_old_pt])

        W_new, n_new_pt = \
            _R(delta, sg.nums_actions, sg.payoff_arrays,
               best_dev_gains, hull.points, hull.vertices,
               hull.equations, u, IC, action_profile_payoff,
               extended_payoff, new_pts, W_new)

        n_iter += 1
        if n_iter >= max_iter:
            break

        # check convergence
        if n_new_pt == n_old_pt:
            if np.linalg.norm(W_new[:n_new_pt] - W_old[:n_new_pt]) < tol:
                break

        # update threat points
        _update_u(u, W_new[:n_new_pt])

    hull = ConvexHull(W_new[:n_new_pt])

    return hull


def _best_dev_gains(rpg):
    """
    Calculate the normalized payoff gains from deviating from the current
    action to the best response for each player.

    Parameters
    ----------
    rpg : RepeatedGame
        Two player repeated game.

    Returns
    -------
    best_dev_gains : tuple(ndarray(float, ndim=2))
        The normalized best deviation payoff gain arrays.
        best_dev_gains[i][ai, a-i] is normalized payoff gain
        player i can get if originally players are choosing
        ai and a-i, and player i deviates to the best response action.
    """
    sg, delta = rpg.sg, rpg.delta

    best_dev_gains = ((1-delta)/delta *
                      (np.max(sg.payoff_arrays[i], 0) - sg.payoff_arrays[i])
                      for i in range(2))

    return tuple(best_dev_gains)


@njit
def _R(delta, nums_actions, payoff_arrays, best_dev_gains, points,
       vertices, equations, u, IC, action_profile_payoff,
       extended_payoff, new_pts, W_new, tol=1e-10):
    """
    Updating the payoff convex hull by iterating all action pairs.
    Using the R operator proposed by Abreu and Sannikov 2014.

    Parameters
    ----------
    delta : scalar(float)
            The common discount rate at which all players discount
            the future.

    nums_actions : tuple(int)
        Tuple of the numbers of actions, one for each player.

    payoff_arrays : tuple(ndarray(float, ndim=2))
        Tuple of the payoff arrays, one for each player.

    best_dev_gains : tuple(ndarray(float, ndim=2))
        Tuple of the normalized best deviation payoff gain arrays.
        best_dev_gains[i][ai, a-i] is payoff gain player i
        can get if originally players are choosing ai and a-i,
        and player i deviates to the best response action.

    points : ndarray(float, ndim=2)
        Coordinates of the points in the W, which construct a
        feasible payoff convex hull.

    vertices : ndarray(float, ndim=1)
        Indices of points forming the vertices of the convex hull,
        which are in counterclockwise order.

    equations : ndarray(float, ndim=2)
        [normal, offset] forming the hyperplane equation of the facet

    u : ndarray(float, ndim=1)
        The threat points.

    IC : ndarray(float, ndim=1)
        The minimum IC continuation values.

    action_profile_payoff : ndarray(float, ndim=1)
        Array of payoff for one action profile.

    extended_payoff : ndarray(float, ndim=2)
        The array [payoff0, payoff1, 1] for checking if
        [payoff0, payoff1] is in the feasible payoff convex hull.

    new_pts : ndarray(float, ndim=1)
        The 4 by 2 array for storing the generated potential
        extreme points of one action profile. One action profile
        can only generate at most 4 points.

    W_new : ndarray(float, ndim=2)
        Array for storing the coordinates of the generated potential
        extreme points that construct a new feasible payoff convex hull.

    tol: scalar(float), optional(default=1e-10)
        The tolerance for checking if two values are equal.

    Returns
    -------
    W_new : ndarray(float, ndim=2)
        The coordinates of the generated potential extreme points
        that construct a new feasible payoff convex hull.

    n_new_pt : scalar(int)
        The number of points in W_new that construct the feasible
        payoff convex hull.
    """
    n_new_pt = 0
    for a0 in range(nums_actions[0]):
        for a1 in range(nums_actions[1]):
            action_profile_payoff[0] = payoff_arrays[0][a0, a1]
            action_profile_payoff[1] = payoff_arrays[1][a1, a0]
            IC[0] = u[0] + best_dev_gains[0][a0, a1]
            IC[1] = u[1] + best_dev_gains[1][a1, a0]

            # check if payoff is larger than IC
            if (action_profile_payoff >= IC).all():
                # check if payoff is inside the convex hull
                extended_payoff[:2] = action_profile_payoff
                if (np.dot(equations, extended_payoff) <= tol).all():
                    W_new[n_new_pt] = action_profile_payoff
                    n_new_pt += 1
                    continue

            new_pts, n = _find_C(new_pts, points, vertices, equations,
                                 extended_payoff, IC, tol)

            for i in range(n):
                W_new[n_new_pt] = \
                    delta * new_pts[i] + (1-delta) * action_profile_payoff
                n_new_pt += 1

    return W_new, n_new_pt


@njit
def _find_C(C, points, vertices, equations, extended_payoff, IC, tol):
    """
    Find all the intersection points between the current convex hull
    and the two IC constraints. It is done by iterating simplex
    counterclockwise.

    Parameters
    ----------
    C : ndarray(float, ndim=2)
        The 4 by 2 array for storing the generated potential
        extreme points of one action profile. One action profile
        can only generate at most 4 points.

    points : ndarray(float, ndim=2)
        Coordinates of the points in the W, which construct a
        feasible payoff convex hull.

    vertices : ndarray(float, ndim=1)
        Indices of points forming the vertices of the convex hull,
        which are in counterclockwise order.

    equations : ndarray(float, ndim=2)
        [normal, offset] forming the hyperplane equation of the facet

    extended_payoff : ndarray(float, ndim=1)
        The array [payoff0, payoff1, 1] for checking if
        [payoff0, payoff1] is in the feasible payoff convex hull.

    IC : ndarray(float, ndim=1)
        The minimum IC continuation values.

    tol : scalar(float)
        The tolerance for checking if two values are equal.

    Returns
    -------
    C : ndarray(float, ndim=2)
        The generated potential extreme points.

    n : scalar(int)
        The number of found intersection points.
    """
    n = 0
    weights = np.empty(2)
    # vertices is ordered counterclockwise
    for i in range(len(vertices)-1):
        n = _intersect(C, n, weights, IC,
                       points[vertices[i]],
                       points[vertices[i+1]], tol)

    n = _intersect(C, n, weights, IC,
                   points[vertices[-1]],
                   points[vertices[0]], tol)

    # check the case that IC is an interior point of the convex hull
    extended_payoff[:2] = IC
    if (np.dot(equations, extended_payoff) <= tol).all():
        C[n, :] = IC
        n += 1

    return C, n


@njit
def _intersect(C, n, weights, IC, pt0, pt1, tol):
    """
    Find the intersection points of a half-closed simplex
    (pt0, pt1] and IC constraints.

    Parameters
    ----------
    C : ndarray(float, ndim=2)
        The 4 by 2 array for storing the generated points of
        one action profile. One action profile can only
        generate at most 4 points.

    n : scalar(int)
        The number of intersection points that have been found.

    weights : ndarray(float, ndim=1)
        The size 2 array for storing the weights when calculate
        the intersection point of simplex and IC constraints.

    IC : ndarray(float, ndim=1)
        The minimum IC continuation values.

    pt0 : ndarray(float, ndim=1)
        Coordinates of the starting point of the simplex.

    pt1 : ndarray(float, ndim=1)
        Coordinates of the ending point of the simplex.

    tol : scalar(float)
        The tolerance for checking if two values are equal.

    Returns
    -------
    n : scalar(int)
        The updated number of found intersection points.
    """
    for i in range(2):
        if (abs(pt0[i] - pt1[i]) < tol):
            if (abs(pt1[i] - IC[i]) < tol):
                x = pt1[1-i]
            else:
                continue
        else:
            weights[i] = (pt0[i] - IC[i]) / (pt0[i] - pt1[i])
            # pt0 is not included to avoid duplication
            # weights in (0, 1]
            if (0 < weights[i] <= 1):
                x = (1 - weights[i]) * pt0[1-i] + weights[i] * pt1[1-i]
            else:
                continue
        # x has to be strictly higher than IC[1-j]
        # if it is equal, then it means IC is one of the vertex
        # it will be added to C in below
        if x - IC[1-i] > tol:
            C[n, i] = IC[i]
            C[n, 1-i] = x
            n += 1
        elif x - IC[1-i] > -tol:
            # to avoid duplication when IC is a vertex
            break

    return n


@njit
def _update_u(u, W):
    """
    Update the threat points if it not feasible in the new W,
    by the minimum of new feasible payoffs.

    Parameters
    ----------
    u : ndarray(float, ndim=1)
        The threat points.

    W : ndarray(float, ndim=1)
        The points that construct the feasible payoff convex hull.

    Returns
    -------
    u : ndarray(float, ndim=1)
        The updated threat points.
    """
    for i in range(2):
        W_min = W[:, i].min()
        if u[i] < W_min:
            u[i] = W_min

    return u



"""
Computes quantities associated with the Gaussian linear state space model.

"""

from textwrap import dedent
import numpy as np
from scipy.linalg import solve
from ._matrix_eqn import solve_discrete_lyapunov
from numba import jit
from .util import check_random_state


@jit(nopython=True)
def simulate_linear_model(A, x0, v, ts_length):
    r"""
    This is a separate function for simulating a vector linear system of
    the form

    .. math::

        x_{t+1} = A x_t + v_t

    given :math:`x_0` = x0

    Here :math:`x_t` and :math:`v_t` are both n x 1 and :math:`A` is n x n.

    The purpose of separating this functionality out is to target it for
    optimization by Numba.  For the same reason, matrix multiplication is
    broken down into for loops.

    Parameters
    ----------
    A : array_like or scalar(float)
        Should be n x n
    x0 : array_like
        Should be n x 1.  Initial condition
    v : np.ndarray
        Should be n x ts_length-1.  Its t-th column is used as the time t
        shock :math:`v_t`
    ts_length : int
        The length of the time series

    Returns
    -------
    x : np.ndarray
        Time series with ts_length columns, the t-th column being :math:`x_t`
    """
    A = np.asarray(A)
    n = A.shape[0]
    x = np.empty((n, ts_length))
    x[:, 0] = x0
    for t in range(ts_length-1):
        # x[:, t+1] = A.dot(x[:, t]) + v[:, t]
        for i in range(n):
            x[i, t+1] = v[i, t]                   # Shock
            for j in range(n):
                x[i, t+1] += A[i, j] * x[j, t]   # Dot Product
    return x


class LinearStateSpace:
    r"""
    A class that describes a Gaussian linear state space model of the
    form:

    .. math::

      x_{t+1} = A x_t + C w_{t+1}

      y_t = G x_t + H v_t

    where :math:`{w_t}` and :math:`{v_t}` are independent and standard normal
    with dimensions k and l respectively.  The initial conditions are
    :math:`\mu_0` and :math:`\Sigma_0` for :math:`x_0 \sim N(\mu_0, \Sigma_0)`.
    When :math:`\Sigma_0=0`, the draw of :math:`x_0` is exactly :math:`\mu_0`.

    Parameters
    ----------
    A : array_like or scalar(float)
        Part of the state transition equation.  It should be `n x n`
    C : array_like or scalar(float)
        Part of the state transition equation.  It should be `n x m`
    G : array_like or scalar(float)
        Part of the observation equation.  It should be `k x n`
    H : array_like or scalar(float), optional(default=None)
        Part of the observation equation.  It should be `k x l`
    mu_0 : array_like or scalar(float), optional(default=None)
        This is the mean of initial draw and is `n x 1`
    Sigma_0 : array_like or scalar(float), optional(default=None)
        This is the variance of the initial draw and is `n x n` and
        also should be positive definite and symmetric

    Attributes
    ----------
    A, C, G, H, mu_0, Sigma_0 : see Parameters
    n, k, m, l : scalar(int)
        The dimensions of x_t, y_t, w_t and v_t respectively

    """

    def __init__(self, A, C, G, H=None, mu_0=None, Sigma_0=None):
        self.A, self.G, self.C = list(map(self.convert, (A, G, C)))
        # = Check Input Shapes = #
        ni, nj = self.A.shape
        if ni != nj:
            raise ValueError(
                "Matrix A (shape: %s) needs to be square" % (self.A.shape, ))
        if ni != self.C.shape[0]:
            raise ValueError(
                "Matrix C (shape: %s) does not have compatible dimensions "
                "with A. It should be shape: %s" % (self.C.shape, (ni, 1)))
        self.m = self.C.shape[1]
        self.k, self.n = self.G.shape
        if self.n != ni:
            raise ValueError("Matrix G (shape: %s) does not have compatible"
                             "dimensions with A (%s)" % (self.G.shape,
                                                         self.A.shape))
        if H is None:
            self.H = None
            self.l = None
        else:
            self.H = self.convert(H)
            self.l = self.H.shape[1]
        if mu_0 is None:
            self.mu_0 = np.zeros((self.n, 1))
        else:
            self.mu_0 = self.convert(mu_0)
            self.mu_0.shape = self.n, 1
        if Sigma_0 is None:
            self.Sigma_0 = np.zeros((self.n, self.n))
        else:
            self.Sigma_0 = self.convert(Sigma_0)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        m = """\
        Linear Gaussian state space model:
          - dimension of state space          : {n}
          - number of innovations             : {m}
          - dimension of observation equation : {k}
        """
        return dedent(m.format(n=self.n, k=self.k, m=self.m))

    def convert(self, x):
        """
        Convert array_like objects (lists of lists, floats, etc.) into
        well formed 2D NumPy arrays

        """
        return np.atleast_2d(np.asarray(x, dtype='float'))

    def simulate(self, ts_length=100, random_state=None):
        r"""
        Simulate a time series of length ts_length, first drawing

        .. math::

            x_0 \sim N(\mu_0, \Sigma_0)

        Parameters
        ----------
        ts_length : scalar(int), optional(default=100)
            The length of the simulation
        random_state : int or np.random.RandomState/Generator, optional
            Random seed (integer) or np.random.RandomState or Generator
            instance to set the initial state of the random number
            generator for reproducibility. If None, a randomly
            initialized RandomState is used.

        Returns
        -------
        x : array_like(float)
            An n x ts_length array, where the t-th column is :math:`x_t`
        y : array_like(float)
            A k x ts_length array, where the t-th column is :math:`y_t`

        """
        random_state = check_random_state(random_state)

        x0 = random_state.multivariate_normal(self.mu_0.flatten(),
                                              self.Sigma_0)
        w = random_state.standard_normal((self.m, ts_length-1))
        v = self.C.dot(w)  # Multiply each w_t by C to get v_t = C w_t
        # == simulate time series == #
        x = simulate_linear_model(self.A, x0, v, ts_length)

        if self.H is not None:
            v = random_state.standard_normal((self.l, ts_length))
            y = self.G.dot(x) + self.H.dot(v)
        else:
            y = self.G.dot(x)

        return x, y

    def replicate(self, T=10, num_reps=100, random_state=None):
        r"""
        Simulate num_reps observations of :math:`x_T` and :math:`y_T` given
        :math:`x_0 \sim N(\mu_0, \Sigma_0)`.

        Parameters
        ----------
        T : scalar(int), optional(default=10)
            The period that we want to replicate values for
        num_reps : scalar(int), optional(default=100)
            The number of replications that we want
        random_state : int or np.random.RandomState/Generator, optional
            Random seed (integer) or np.random.RandomState or Generator
            instance to set the initial state of the random number
            generator for reproducibility. If None, a randomly
            initialized RandomState is used.

        Returns
        -------
        x : array_like(float)
            An n x num_reps array, where the j-th column is the j_th
            observation of :math:`x_T`

        y : array_like(float)
            A k x num_reps array, where the j-th column is the j_th
            observation of :math:`y_T`

        """
        random_state = check_random_state(random_state)

        x = np.empty((self.n, num_reps))
        for j in range(num_reps):
            x_T, _ = self.simulate(ts_length=T+1, random_state=random_state)
            x[:, j] = x_T[:, -1]
        if self.H is not None:
            v = random_state.standard_normal((self.l, num_reps))
            y = self.G.dot(x) + self.H.dot(v)
        else:
            y = self.G.dot(x)

        return x, y

    def moment_sequence(self):
        r"""
        Create a generator to calculate the population mean and
        variance-covariance matrix for both :math:`x_t` and :math:`y_t`
        starting at the initial condition (self.mu_0, self.Sigma_0).
        Each iteration produces a 4-tuple of items (mu_x, mu_y, Sigma_x,
        Sigma_y) for the next period.

        Yields
        ------
        mu_x : array_like(float)
            An n x 1 array representing the population mean of x_t
        mu_y : array_like(float)
            A  k x 1 array representing the population mean of y_t
        Sigma_x : array_like(float)
            An n x n array representing the variance-covariance matrix
            of x_t
        Sigma_y : array_like(float)
            A k x k array representing the variance-covariance matrix
            of y_t

        """
        # == Simplify names == #
        A, C, G, H = self.A, self.C, self.G, self.H
        # == Initial moments == #
        mu_x, Sigma_x = self.mu_0, self.Sigma_0

        while 1:
            mu_y = G.dot(mu_x)
            if H is None:
                Sigma_y = G.dot(Sigma_x).dot(G.T)
            else:
                Sigma_y = G.dot(Sigma_x).dot(G.T) + H.dot(H.T)

            yield mu_x, mu_y, Sigma_x, Sigma_y

            # == Update moments of x == #
            mu_x = A.dot(mu_x)
            Sigma_x = A.dot(Sigma_x).dot(A.T) + C.dot(C.T)

    def stationary_distributions(self):
        r"""
        Compute the moments of the stationary distributions of :math:`x_t` and
        :math:`y_t` if possible. Computation is by solving the discrete
        Lyapunov equation.

        Returns
        -------
        mu_x : array_like(float)
            An n x 1 array representing the stationary mean of :math:`x_t`
        mu_y : array_like(float)
            An k x 1 array representing the stationary mean of :math:`y_t`
        Sigma_x : array_like(float)
            An n x n array representing the stationary var-cov matrix
            of :math:`x_t`
        Sigma_y : array_like(float)
            An k x k array representing the stationary var-cov matrix
            of :math:`y_t`
        Sigma_yx : array_like(float)
            An k x n array representing the stationary cov matrix
            between :math:`y_t` and :math:`x_t`.

        """
        self.__partition()
        num_const, sorted_idx = self.num_const, self.sorted_idx
        A21, A22 = self.A21, self.A22
        CC2 = self.C2 @ self.C2.T
        n = self.n

        if num_const > 0:
            μ = solve(np.eye(n-num_const) - A22, A21)
        else:
            μ = solve(np.eye(n-num_const) - A22, np.zeros((n, 1)))
        Σ = solve_discrete_lyapunov(A22, CC2, method='bartels-stewart')

        mu_x = np.empty((n, 1))
        mu_x[:num_const] = self.mu_0[sorted_idx[:num_const]]
        mu_x[num_const:] = μ

        Sigma_x = np.zeros((n, n))
        Sigma_x[num_const:, num_const:] = Σ

        mu_x = self.P.T @ mu_x
        Sigma_x = self.P.T @ Sigma_x @ self.P

        mu_y = self.G @ mu_x
        Sigma_y = self.G @ Sigma_x @ self.G.T
        if self.H is not None:
            Sigma_y += self.H @ self.H.T
        Sigma_yx = self.G @ Sigma_x

        self.mu_x, self.mu_y = mu_x, mu_y
        self.Sigma_x, self.Sigma_y, self.Sigma_yx = Sigma_x, Sigma_y, Sigma_yx

        return mu_x, mu_y, Sigma_x, Sigma_y, Sigma_yx

    def geometric_sums(self, beta, x_t):
        r"""
        Forecast the geometric sums

        .. math::

            S_x := E \Big[ \sum_{j=0}^{\infty} \beta^j x_{t+j} | x_t \Big]

            S_y := E \Big[ \sum_{j=0}^{\infty} \beta^j y_{t+j} | x_t \Big]

        Parameters
        ----------
        beta : scalar(float)
            Discount factor, in [0, 1)

        beta : array_like(float)
            The term x_t for conditioning

        Returns
        -------
        S_x : array_like(float)
            Geometric sum as defined above

        S_y : array_like(float)
            Geometric sum as defined above

        """

        I = np.identity(self.n)
        S_x = solve(I - beta * self.A, x_t)
        S_y = self.G.dot(S_x)

        return S_x, S_y

    def impulse_response(self, j=5):
        r"""
        Pulls off the imuplse response coefficients to a shock
        in :math:`w_{t}` for :math:`x` and :math:`y`

        Important to note: We are uninterested in the shocks to
        v for this method

        * :math:`x` coefficients are :math:`C, AC, A^2 C...`
        * :math:`y` coefficients are :math:`GC, GAC, GA^2C...`

        Parameters
        ----------
        j : Scalar(int)
            Number of coefficients that we want

        Returns
        -------
        xcoef : list(array_like(float, 2))
            The coefficients for x
        ycoef : list(array_like(float, 2))
            The coefficients for y
        """
        # Pull out matrices
        A, C, G, H = self.A, self.C, self.G, self.H
        Apower = np.copy(A)

        # Create room for coefficients
        xcoef = [C]
        ycoef = [np.dot(G, C)]

        for i in range(j):
            xcoef.append(np.dot(Apower, C))
            ycoef.append(np.dot(G, np.dot(Apower, C)))
            Apower = np.dot(Apower, A)

        return xcoef, ycoef

    def __partition(self):
        r"""
        Reorder the states by shifting the constant terms to the
        top of the state vector. Then partition the linear state
        space model following Appendix C in RMT Ch2 such that the
        A22 matrix associated with non-constant states have eigenvalues
        all strictly smaller than 1.

        .. math::
            \left[\begin{array}{c}
            const\\
            x_{2,t+1}
            \end{array}\right]=\left[\begin{array}{cc}
            I & 0\\
            A_{21} & A_{22}
            \end{array}\right]\left[\begin{array}{c}
            1\\
            x_{2,t}
            \end{array}\right]+\left[\begin{array}{c}
            0\\
            C_{2}
            \end{array}\right]w_{t+1}

        Returns
        -------
        A22 : array_like(float)
            Lower right part of the reordered matrix A.
        A21 : array_like(float)
            Lower left part of the reordered matrix A.
        """
        A, C = self.A, self.C
        n = self.n

        sorted_idx = []
        A_diag = np.diag(A)
        num_const = 0
        for idx in range(n):
            if (A_diag[idx] == 1) and (C[idx, :] == 0).all() and \
                    np.linalg.norm(A[idx, :]) == 1:
                sorted_idx.insert(0, idx)
                num_const += 1
            else:
                sorted_idx.append(idx)
        self.num_const = num_const
        self.sorted_idx = sorted_idx

        P = np.zeros((n, n))
        P[range(n), sorted_idx] = 1

        sorted_A = P @ A @ P.T
        sorted_C = P @ C
        A21 = sorted_A[num_const:, :num_const]
        A22 = sorted_A[num_const:, num_const:]
        C2 = sorted_C[num_const:, :]

        self.P, self.A21, self.A22, self.C2 = P, A21, A22, C2

        return A21, A22



"""
Tests for util/random.py

Functions
---------
probvec
sample_without_replacement

"""
import numbers
import numpy as np
from numpy.testing import (assert_array_equal, assert_allclose, assert_raises,
                           assert_)
from numba import njit
from quantecon.random import probvec, sample_without_replacement, draw


# probvec #

class TestProbvec:
    def setup_method(self):
        self.m, self.k = 2, 3  # m vectors of dimension k
        seed = 1234

        self.out_parallel = probvec(self.m, self.k, random_state=seed)
        self.out_cpu = \
            probvec(self.m, self.k, random_state=seed, parallel=False)

    def test_shape(self):
        for out in [self.out_parallel, self.out_cpu]:
            assert_(out.shape == (self.m, self.k))

    def test_parallel_cpu(self):
        assert_array_equal(self.out_parallel, self.out_cpu)


# sample_without_replacement #

def test_sample_without_replacement_shape():
    assert_array_equal(sample_without_replacement(2, 0).shape, (0,))

    n, k, m = 5, 3, 4
    assert_array_equal(
        sample_without_replacement(n, k).shape,
        (k,)
    )
    assert_array_equal(
        sample_without_replacement(n, k, num_trials=m).shape,
        (m, k)
    )


def test_sample_without_replacement_uniqueness():
    n = 10
    a = sample_without_replacement(n, n)
    b = np.unique(a)
    assert_(len(b) == n)


def test_sample_without_replacement_value_error():
    # n <= 0
    assert_raises(ValueError, sample_without_replacement, 0, 2)
    assert_raises(ValueError, sample_without_replacement, -1, -1)

    # k > n
    assert_raises(ValueError, sample_without_replacement, 2, 3)


# draw #

@njit
def draw_jitted(cdf, size=None):
    return draw(cdf, size)


class TestDraw:
    def setup_method(self):
        self.pmf = np.array([0.4, 0.1, 0.5])
        self.cdf = np.cumsum(self.pmf)
        self.n = len(self.pmf)
        self.draw_funcs = [draw, draw_jitted]

    def test_return_types(self):
        for func in self.draw_funcs:
            out = func(self.cdf)
            assert_(isinstance(out, numbers.Integral))

        size = 10
        for func in self.draw_funcs:
            out = func(self.cdf, size)
            assert_(out.shape == (size,))

    def test_return_values(self):
        for func in self.draw_funcs:
            out = func(self.cdf)
            assert_(out in range(self.n))

        size = 10
        for func in self.draw_funcs:
            out = func(self.cdf, size)
            assert_(np.isin(out, range(self.n)).all())

    def test_lln(self):
        size = 1000000
        for func in self.draw_funcs:
            out = func(self.cdf, size)
            hist, bin_edges = np.histogram(out, bins=self.n, density=True)
            pmf_computed = hist * np.diff(bin_edges)
            atol = 1e-2
            assert_allclose(pmf_computed, self.pmf, atol=atol)



"""
Tests for markov/ddp.py

"""
import numpy as np
import scipy.sparse as sparse
from numpy.testing import (assert_array_equal, assert_allclose, assert_raises,
                           assert_)

from quantecon.markov import DiscreteDP, backward_induction


class TestDiscreteDP:
    def setup_method(self):
        # From Puterman 2005, Section 3.1
        beta = 0.95

        # Formulation with R: n x m, Q: n x m x n
        n, m = 2, 2  # number of states, number of actions
        R = [[5, 10], [-1, -np.inf]]
        Q = np.empty((n, m, n))
        Q[0, 0, :] = 0.5, 0.5
        Q[0, 1, :] = 0, 1
        Q[1, :, :] = 0, 1
        ddp0 = DiscreteDP(R, Q, beta)

        # Formulation with state-action pairs
        L = 3  # number of state-action pairs
        s_indices = [0, 0, 1]
        a_indices = [0, 1, 0]
        R_sa = [R[0][0], R[0][1], R[1][0]]
        Q_sa = sparse.lil_matrix((L, n))
        Q_sa[0, :] = Q[0, 0, :]
        Q_sa[1, :] = Q[0, 1, :]
        Q_sa[2, :] = Q[1, 0, :]
        ddp_sa_sparse = DiscreteDP(R_sa, Q_sa, beta, s_indices, a_indices)
        ddp_sa_dense = \
            DiscreteDP(R_sa, Q_sa.toarray(), beta, s_indices, a_indices)

        self.ddps = [ddp0, ddp_sa_sparse, ddp_sa_dense]

        for ddp in self.ddps:
            ddp.max_iter = 200

        self.epsilon = 1e-2

        # Analytical solution for beta > 10/11, Example 6.2.1
        self.v_star = [(5-5.5*beta)/((1-0.5*beta)*(1-beta)), -1/(1-beta)]
        self.sigma_star = [0, 0]

    def test_value_iteration(self):
        for ddp in self.ddps:
            res = ddp.solve(method='value_iteration', epsilon=self.epsilon)

            v_init = [0, 0]
            res_init = ddp.solve(method='value_iteration', v_init=v_init,
                                 epsilon=self.epsilon)

            # Check v is an epsilon/2-approximation of v_star
            assert_(np.abs(res.v - self.v_star).max() < self.epsilon/2)
            assert_(np.abs(res_init.v - self.v_star).max() < self.epsilon/2)

            # Check sigma == sigma_star
            assert_array_equal(res.sigma, self.sigma_star)
            assert_array_equal(res_init.sigma, self.sigma_star)

    def test_policy_iteration(self):
        for ddp in self.ddps:
            res = ddp.solve(method='policy_iteration')

            v_init = [0, 1]
            res_init = ddp.solve(method='policy_iteration', v_init=v_init)

            # Check v == v_star
            assert_allclose(res.v, self.v_star)
            assert_allclose(res_init.v, self.v_star)

            # Check sigma == sigma_star
            assert_array_equal(res.sigma, self.sigma_star)
            assert_array_equal(res_init.sigma, self.sigma_star)

    def test_modified_policy_iteration(self):
        for ddp in self.ddps:
            res = ddp.solve(method='modified_policy_iteration',
                            epsilon=self.epsilon)

            v_init = [0, 1]
            res_init = ddp.solve(method='modified_policy_iteration',
                                 v_init=v_init,
                                 epsilon=self.epsilon)

            # Check v is an epsilon/2-approximation of v_star
            assert_(np.abs(res.v - self.v_star).max() < self.epsilon/2)
            assert_(np.abs(res_init.v - self.v_star).max() < self.epsilon/2)

            # Check sigma == sigma_star
            assert_array_equal(res.sigma, self.sigma_star)
            assert_array_equal(res_init.sigma, self.sigma_star)

    def test_modified_policy_iteration_k0(self):
        k = 0
        for ddp in self.ddps:
            res = ddp.solve(method='modified_policy_iteration',
                            epsilon=self.epsilon, k=k)

            # Check v is an epsilon/2-approximation of v_star
            assert_(np.abs(res.v - self.v_star).max() < self.epsilon/2)

            # Check sigma == sigma_star
            assert_array_equal(res.sigma, self.sigma_star)

    def test_linear_programming(self):
        for ddp in self.ddps:
            if ddp._sparse:
                assert_raises(NotImplementedError, ddp.solve,
                              method='linear_programming')
            else:
                res = ddp.solve(method='linear_programming')

                v_init = [0, 1]
                res_init = ddp.solve(method='linear_programming',
                                     v_init=v_init)

                # Check v == v_star
                assert_allclose(res.v, self.v_star)
                assert_allclose(res_init.v, self.v_star)

                # Check sigma == sigma_star
                assert_array_equal(res.sigma, self.sigma_star)
                assert_array_equal(res_init.sigma, self.sigma_star)


def test_ddp_beta_0():
    n, m = 3, 2
    R = np.array([[0, 1], [1, 0], [0, 1]])
    Q = np.empty((n, m, n))
    Q[:] = 1/n
    beta = 0
    sigma_star = [1, 0, 1]
    v_star = [1, 1, 1]
    v_init = [0, 0, 0]

    ddp0 = DiscreteDP(R, Q, beta)
    ddp1 = ddp0.to_sa_pair_form()
    ddp2 = ddp0.to_sa_pair_form(sparse=False)
    methods = ['vi', 'pi', 'mpi', 'lp']

    for ddp in [ddp0, ddp1, ddp2]:
        for method in methods:
            if method == 'lp' and ddp._sparse:
                assert_raises(NotImplementedError, ddp.solve,
                              method=method, v_init=v_init)
            else:
                res = ddp.solve(method=method, v_init=v_init)
                assert_array_equal(res.sigma, sigma_star)
                assert_array_equal(res.v, v_star)


def test_ddp_sorting():
    beta = 0.95

    # Sorted
    s_indices = [0, 0, 1]
    a_indices = [0, 1, 0]
    a_indptr = [0, 2, 3]
    R = [0, 1, 2]
    Q = [(1, 0), (1/2, 1/2), (0, 1)]
    Q_sparse = sparse.csr_matrix(Q)

    # Shuffled
    s_indices_shuffled = [0, 1, 0]
    a_indices_shuffled = [0, 0, 1]
    R_shuffled = [0, 2, 1]
    Q_shuffled = [(1, 0), (0, 1), (1/2, 1/2)]
    Q_shuffled_sparse = sparse.csr_matrix(Q_shuffled)

    ddp0 = DiscreteDP(R, Q, beta, s_indices, a_indices)
    ddp_sparse = DiscreteDP(R, Q_sparse, beta, s_indices, a_indices)
    ddp_shuffled = DiscreteDP(R_shuffled, Q_shuffled, beta,
                              s_indices_shuffled, a_indices_shuffled)
    ddp_shuffled_sparse = DiscreteDP(R_shuffled, Q_shuffled_sparse, beta,
                                     s_indices_shuffled, a_indices_shuffled)

    for ddp in [ddp0, ddp_sparse, ddp_shuffled, ddp_shuffled_sparse]:
        assert_array_equal(ddp.s_indices, s_indices)
        assert_array_equal(ddp.a_indices, a_indices)
        assert_array_equal(ddp.a_indptr, a_indptr)
        assert_array_equal(ddp.R, R)
        if sparse.issparse(ddp.Q):
            ddp_Q = ddp.Q.toarray()
        else:
            ddp_Q = ddp.Q
        assert_array_equal(ddp_Q, Q)


class TestFiniteHorizon:
    def setup_method(self):
        # From Puterman 2005, Section 3.2, Section 4.6.1
        # "single-product stochastic inventory control"
        s_indices = [0, 0, 0, 0, 1, 1, 1, 2, 2, 3]
        a_indices = [0, 1, 2, 3, 0, 1, 2, 0, 1, 0]
        R = [ 0., -1., -2., -5.,  5.,  0., -3.,  6., -1.,  5.]
        Q = [[ 1.  ,  0.  ,  0.  ,  0.  ],
             [ 0.75,  0.25,  0.  ,  0.  ],
             [ 0.25,  0.5 ,  0.25,  0.  ],
             [ 0.  ,  0.25,  0.5 ,  0.25],
             [ 0.75,  0.25,  0.  ,  0.  ],
             [ 0.25,  0.5 ,  0.25,  0.  ],
             [ 0.  ,  0.25,  0.5 ,  0.25],
             [ 0.25,  0.5 ,  0.25,  0.  ],
             [ 0.  ,  0.25,  0.5 ,  0.25],
             [ 0.  ,  0.25,  0.5 ,  0.25]]
        beta = 1
        self.ddp = DiscreteDP(R, Q, beta, s_indices, a_indices)

    def test_backward_induction(self):
        T = 3
        # v_T = np.zeors(self.ddp.n)
        vs_expected = [[67/16, 129/16, 194/16, 227/16],
                       [2, 25/4, 10, 21/2],
                       [0, 5, 6, 5],
                       [0, 0, 0, 0]]
        sigmas_expected = [[3, 0, 0, 0],
                           [2, 0, 0, 0],
                           [0, 0, 0, 0]]

        vs, sigmas = backward_induction(self.ddp, T)

        assert_allclose(vs, vs_expected)
        assert_array_equal(sigmas, sigmas_expected)


def test_ddp_negative_inf_error():
    n, m = 3, 2
    R = np.array([[0, 1], [0, -np.inf], [-np.inf, -np.inf]])
    Q = np.empty((n, m, n))
    Q[:] = 1/n

    s_indices = [0, 0, 1, 1, 2, 2]
    a_indices = [0, 1, 0, 1, 0, 1]
    R_sa = R.reshape(n*m)
    Q_sa_dense = Q.reshape(n*m, n)
    Q_sa_sparse = sparse.csr_matrix(Q_sa_dense)

    beta = 0.95

    assert_raises(ValueError, DiscreteDP, R, Q, beta)
    assert_raises(
        ValueError, DiscreteDP, R_sa, Q_sa_dense, beta, s_indices, a_indices
    )
    assert_raises(
        ValueError, DiscreteDP, R_sa, Q_sa_sparse, beta, s_indices, a_indices
    )


def test_ddp_no_feasibile_action_error():
    # No action is feasible at state 1
    s_indices = [0, 0, 2, 2]
    a_indices = [0, 1, 0, 1]
    R = [1, 0, 0, 1]
    Q = [(1/3, 1/3, 1/3) for i in range(4)]
    beta = 0.95

    assert_raises(ValueError, DiscreteDP, R, Q, beta, s_indices, a_indices)


def test_ddp_beta_1_not_implemented_error():
    n, m = 3, 2
    R = np.array([[0, 1], [1, 0], [0, 1]])
    Q = np.empty((n, m, n))
    Q[:] = 1/n
    beta = 1

    ddp0 = DiscreteDP(R, Q, beta)
    ddp1 = ddp0.to_sa_pair_form()
    ddp2 = ddp0.to_sa_pair_form(sparse=False)

    solution_methods = \
        ['value_iteration', 'policy_iteration', 'modified_policy_iteration']

    for ddp in [ddp0, ddp1, ddp2]:
        assert_raises(NotImplementedError, ddp.evaluate_policy, np.zeros(n))
        for method in solution_methods:
            assert_raises(NotImplementedError, getattr(ddp, method))


def test_ddp_to_sa_and_to_product():
    n, m = 3, 2
    R = np.array([[0, 1], [1, 0], [-np.inf, 1]])
    Q = np.empty((n, m, n))
    Q[:] = 1/n
    Q[0, 0, 0] = 0
    Q[0, 0, 1] = 2/n
    beta = 0.95

    sparse_R = np.array([0, 1, 1, 0, 1])
    _Q = np.full((5, 3), 1/3)
    _Q[0, 0] = 0
    _Q[0, 1] = 2/n
    sparse_Q = sparse.coo_matrix(_Q)

    ddp = DiscreteDP(R, Q, beta)
    ddp_sa = ddp.to_sa_pair_form()
    ddp_sa2 = ddp_sa.to_sa_pair_form()
    ddp_sa3 = ddp.to_sa_pair_form(sparse=False)
    ddp2 = ddp_sa.to_product_form()
    ddp3 = ddp_sa2.to_product_form()
    ddp4 = ddp.to_product_form()

    # make sure conversion worked
    for ddp_s in [ddp_sa, ddp_sa2, ddp_sa3]:
        assert_allclose(ddp_s.R, sparse_R)
        # allcose doesn't work on sparse
        np.max(np.abs((sparse_Q - ddp_s.Q))) < 1e-15
        assert_allclose(ddp_s.beta, beta)

    # these two will have probability 0 in state 2, action 0 b/c
    # of the infeasiability in R
    funky_Q = np.empty((n, m, n))
    funky_Q[:] = 1/n
    funky_Q[0, 0, 0] = 0
    funky_Q[0, 0, 1] = 2/n
    funky_Q[2, 0, :] = 0
    for ddp_f in [ddp2, ddp3]:
        assert_allclose(ddp_f.R, ddp.R)
        assert_allclose(ddp_f.Q, funky_Q)
        assert_allclose(ddp_f.beta, ddp.beta)

    # this one is just the original one.
    assert_allclose(ddp4.R, ddp.R)
    assert_allclose(ddp4.Q, ddp.Q)
    assert_allclose(ddp4.beta, ddp.beta)

    for method in ["pi", "vi", "mpi"]:
        sol1 = ddp.solve(method=method)
        for ddp_other in [ddp_sa, ddp_sa2, ddp_sa3, ddp2, ddp3, ddp4]:
            sol2 = ddp_other.solve(method=method)

            for k in ["v", "sigma", "num_iter"]:
                assert_allclose(sol1[k], sol2[k])



"""
Tests for lcp_lemke

"""
import numpy as np
from numpy.testing import assert_, assert_allclose, assert_equal

from quantecon.optimize import lcp_lemke


def _assert_ray_termination(res):
    # res: lcp result object
    assert_(not res.success, "incorrectly reported success")
    assert_equal(res.status, 2, "failed to report ray termination status")


def _assert_success(res, M, q, desired_z=None, rtol=1e-15, atol=1e-15):
    if not res.success:
        msg = "lcp_lemke status {0}".format(res.status)
        raise AssertionError(msg)

    assert_equal(res.status, 0)
    if desired_z is not None:
        assert_allclose(res.z, desired_z,
                        err_msg="converged to an unexpected solution",
                        rtol=rtol, atol=atol)

    assert_((res.z >= -atol).all())

    w = M @ res.z + q
    assert_((w >= -atol).all())

    assert_allclose(w * res.z, np.zeros_like(res.z), rtol=rtol, atol=atol)


class TestLCPLemke:
    def test_Murty_Ex_2_8(self):
        M = [[1, -1, -1, -1],
             [-1, 1, -1, -1],
             [1, 1, 2, 0],
             [1, 1, 0, 2]]
        q = [3, 5, -9, -5]
        M, q = map(np.asarray, [M, q])
        res = lcp_lemke(M, q)
        _assert_success(res, M, q)

    def test_Murty_Ex_2_9(self):
        M = [[-1, 0, -3],
             [1, -2, -5],
             [-2, -1, -2]]
        q = [-3, -2, -1]
        M, q = map(np.asarray, [M, q])
        res = lcp_lemke(M, q)
        _assert_ray_termination(res)

    def test_Kostreva_Ex_1(self):
        # Cycling without careful tie breaking
        M = [[1, 2, 0],
             [0, 1, 2],
             [2, 0, 1]]
        q = [-1, -1, -1]
        M, q = map(np.asarray, [M, q])
        res = lcp_lemke(M, q)
        _assert_success(res, M, q)

    def test_Kostreva_Ex_2(self):
        # Cycling without careful tie breaking
        M = [[1, -1, 3],
             [2, -1, 3],
             [-1, -2, 0]]
        q = [-1, -1, -1]
        M, q = map(np.asarray, [M, q])
        res = lcp_lemke(M, q)
        _assert_ray_termination(res)

    def test_Murty_Ex_2_11(self):
        M = [[-1.5, 2],
             [-4, 4]]
        q = [-5, 17]
        d = [5., 16.]
        M, q, d = map(np.asarray, [M, q, d])
        res = lcp_lemke(M, q, d=d)
        _assert_ray_termination(res)

        res = lcp_lemke(M, q, d=np.ones_like(d))
        _assert_success(res, M, q, atol=1e-13)

    def test_bimatrix_game(self):
        A = [[3, 3],
             [2, 5],
             [0, 6]]
        B = [[3, 2, 3],
             [2, 6, 1]]
        A, B = map(np.asarray, [A, B])
        m, n = A.shape
        I = np.cumsum([0, m, n, m, m, n, n])
        M = np.zeros((3*m+3*n, 3*m+3*n))
        M[I[0]:I[1], I[1]:I[2]] = -A + A.max()
        M[I[0]:I[1], I[2]:I[3]], M[I[0]:I[1], I[3]:I[4]] = 1, -1
        M[I[1]:I[2], I[0]:I[1]] = -B + B.max()
        M[I[1]:I[2], I[4]:I[5]], M[I[1]:I[2], I[5]:I[6]] = 1, -1
        M[I[2]:I[3], I[0]:I[1]], M[I[3]:I[4], I[0]:I[1]] = -1, 1
        M[I[4]:I[5], I[1]:I[2]], M[I[5]:I[6], I[1]:I[2]] = -1, 1
        q = np.zeros(3*m+3*n)
        q[I[2]:I[3]], q[I[3]:I[4]] = 1, -1
        q[I[4]:I[5]], q[I[5]:I[6]] = 1, -1

        res = lcp_lemke(M, q)
        _assert_success(res, M, q)



"""
Filename: test_logitdyn.py

Tests for logitdyn.py

"""
import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal_nulp

from quantecon.game_theory import NormalFormGame, LogitDynamics


class TestLogitDynamics:
    '''Test the methods of LogitDynamics'''

    def setup_method(self):
        '''Setup a LogitDynamics instance'''
        # symmetric 2x2 coordination game
        payoff_matrix = [[4, 0],
                         [3, 2]]
        beta = 4.0
        data = NormalFormGame(payoff_matrix)
        self.ld = LogitDynamics(data, beta=beta)

    def test_play(self):
        seed = 76240103339929371127372784081282227092
        x = [self.ld.play(init_actions=(0, 0),
                          random_state=np.random.default_rng(seed))
             for i in range(2)]
        assert_array_equal(x[0], x[1])

    def test_time_series(self):
        seed = 165570719993771384215214311194249493239
        series = [self.ld.time_series(ts_length=10, init_actions=(0, 0),
                                      random_state=np.random.default_rng(seed))
                  for i in range(2)]
        assert_array_equal(series[0], series[1])


def test_set_choice_probs_with_asymmetric_payoff_matrix():
    bimatrix = np.array([[(4, 4), (1, 1), (0, 3)],
                         [(3, 0), (1, 1), (2, 2)]])
    beta = 1.0
    data = NormalFormGame(bimatrix)
    ld = LogitDynamics(data, beta=beta)

    # (Normalized) CDFs of logit choice
    cdfs = np.ones((bimatrix.shape[1], bimatrix.shape[0]))
    cdfs[:, 0] = 1 / (1 + np.exp(beta*(bimatrix[1, :, 0]-bimatrix[0, :, 0])))

    # self.ld.players[0].logit_choice_cdfs: unnormalized
    cdfs_computed = ld.players[0].logit_choice_cdfs
    cdfs_computed = cdfs_computed / cdfs_computed[..., [-1]]  # Normalized

    assert_array_almost_equal_nulp(cdfs_computed, cdfs)



"""
Collection of functions to approximate a continuous random process by a
discrete Markov chain.

"""

from math import erfc, sqrt
from .core import MarkovChain
from .estimate import fit_discrete_mc
from .._lss import simulate_linear_model
from .._matrix_eqn import solve_discrete_lyapunov
from ..util import check_random_state

import warnings
import numpy as np
import numbers
from numba import njit


def rouwenhorst(n, rho, sigma, mu=0.):
    r"""
    Takes as inputs n, mu, sigma, rho. It will then construct a markov chain
    that estimates an AR(1) process of:
    :math:`y_t = \mu + \rho y_{t-1} + \varepsilon_t`
    where :math:`\varepsilon_t` is i.i.d. normal of mean 0, std dev of sigma

    The Rouwenhorst approximation uses the following recursive defintion
    for approximating a distribution:

    .. math::

        \theta_2 =
        \begin{bmatrix}
        p     &  1 - p \\
        1 - q &  q     \\
        \end{bmatrix}

    .. math::

        \theta_{n+1} =
        p
        \begin{bmatrix}
        \theta_n & 0   \\
        0        & 0   \\
        \end{bmatrix}
        + (1 - p)
        \begin{bmatrix}
        0  & \theta_n  \\
        0  &  0        \\
        \end{bmatrix}
        + q
        \begin{bmatrix}
        0        & 0   \\
        \theta_n & 0   \\
        \end{bmatrix}
        + (1 - q)
        \begin{bmatrix}
        0  &  0        \\
        0  & \theta_n  \\
        \end{bmatrix}

    where :math:`{p = q = \frac{(1 + \rho)}{2}}`

    Parameters
    ----------
    n : int
        The number of points to approximate the distribution

    rho : float
        Persistence parameter in AR(1) process, if you are approximating
        an AR(1) process then this is the autocorrelation across periods.

    sigma : float
        The value of the standard deviation of the :math:`\varepsilon` process

    mu : float, optional(default=0.0)
        The value :math:`\mu` in the process.  Note that the mean of this
        AR(1) process, :math:`y`, is simply :math:`\mu/(1 - \rho)`

    Returns
    -------

    mc : MarkovChain
        An instance of the MarkovChain class that stores the transition
        matrix and state values returned by the discretization method

    Notes
    -----

    UserWarning: The API of `rouwenhorst` was changed from
    `rouwenhorst(n, ybar, sigma, rho)` to
    `rouwenhorst(n, rho, sigma, mu=0.)` in version 0.6.0.

    """

    warnings.warn("The API of rouwenhorst has changed from `rouwenhorst(n, ybar, sigma, rho)`"
                  " to `rouwenhorst(n, rho, sigma, mu=0.)`. To find more details please visit:"
                  " https://github.com/QuantEcon/QuantEcon.py/issues/663.",
                  UserWarning, stacklevel=2)
    # Get the standard deviation of y
    y_sd = sqrt(sigma**2 / (1 - rho**2))

    # Given the moments of our process we can find the right values
    # for p, q, psi because there are analytical solutions as shown in
    # Gianluca Violante's notes on computational methods
    p = (1 + rho) / 2
    q = p
    psi = y_sd * np.sqrt(n - 1)

    # Find the states
    ubar = psi
    lbar = -ubar

    bar = np.linspace(lbar, ubar, n)

    def row_build_mat(n, p, q):
        """
        This method uses the values of p and q to build the transition
        matrix for the rouwenhorst method

        """

        if n == 2:
            theta = np.array([[p, 1 - p], [1 - q, q]])

        elif n > 2:
            p1 = np.zeros((n, n))
            p2 = np.zeros((n, n))
            p3 = np.zeros((n, n))
            p4 = np.zeros((n, n))

            new_mat = row_build_mat(n - 1, p, q)

            p1[:n - 1, :n - 1] = p * new_mat
            p2[:n - 1, 1:] = (1 - p) * new_mat
            p3[1:, :-1] = (1 - q) * new_mat
            p4[1:, 1:] = q * new_mat

            theta = p1 + p2 + p3 + p4
            theta[1:n - 1, :] = theta[1:n - 1, :] / 2

        else:
            raise ValueError("The number of states must be positive " +
                             "and greater than or equal to 2")

        return theta

    theta = row_build_mat(n, p, q)

    bar += mu / (1 - rho)

    return MarkovChain(theta, bar)


def tauchen(n, rho, sigma, mu=0., n_std=3):
    r"""
    Computes a Markov chain associated with a discretized version of
    the linear Gaussian AR(1) process

    .. math::

        y_t = \mu + \rho y_{t-1} + \epsilon_t

    using Tauchen's method. Here :math:`{\epsilon_t}` is an i.i.d. Gaussian process
    with zero mean.

    Parameters
    ----------

    n : scalar(int)
        The number of states to use in the approximation
    rho : scalar(float)
        The autocorrelation coefficient, Persistence parameter in AR(1) process
    sigma : scalar(float)
        The standard deviation of the random process
    mu : scalar(float), optional(default=0.0)
        The value :math:`\mu` in the process.  Note that the mean of this
        AR(1) process, :math:`y`, is simply :math:`\mu/(1 - \rho)`
    n_std : scalar(int), optional(default=3)
        The number of standard deviations to approximate out to

    Returns
    -------

    mc : MarkovChain
        An instance of the MarkovChain class that stores the transition
        matrix and state values returned by the discretization method

    Notes
    -----

    UserWarning: The API of `tauchen` was changed from
    `tauchen(rho, sigma_u, b=0., m=3, n=7)` to
    `tauchen(n, rho, sigma, mu=0., n_std=3)` in version 0.6.0.

    """

    if not isinstance(n, numbers.Integral):
        warnings.warn(
            "The API of tauchen has changed from `tauchen(rho, sigma_u, b=0., m=3, n=7)`"
            " to `tauchen(n, rho, sigma, mu=0., n_std=3)`. To find more details please visit:"
            " https://github.com/QuantEcon/QuantEcon.py/issues/663.",
            UserWarning, stacklevel=2)

    # standard deviation of demeaned y_t
    std_y = np.sqrt(sigma**2 / (1 - rho**2))

    # top of discrete state space for demeaned y_t
    x_max = n_std * std_y

    # bottom of discrete state space for demeaned y_t
    x_min = -x_max

    # discretized state space for demeaned y_t
    x = np.linspace(x_min, x_max, n)

    step = (x_max - x_min) / (n - 1)
    half_step = 0.5 * step
    P = np.empty((n, n))

    # approximate Markov transition matrix for
    # demeaned y_t
    _fill_tauchen(x, P, n, rho, sigma, half_step)

    # shifts the state values by the long run mean of y_t
    mu = mu / (1 - rho)

    mc = MarkovChain(P, state_values=x+mu)

    return mc


@njit
def std_norm_cdf(x):
    return 0.5 * erfc(-x / sqrt(2))


@njit
def _fill_tauchen(x, P, n, rho, sigma, half_step):
    for i in range(n):
        P[i, 0] = std_norm_cdf((x[0] - rho * x[i] + half_step) / sigma)
        P[i, n - 1] = 1 - \
            std_norm_cdf((x[n - 1] - rho * x[i] - half_step) / sigma)
        for j in range(1, n - 1):
            z = x[j] - rho * x[i]
            P[i, j] = (std_norm_cdf((z + half_step) / sigma) -
                       std_norm_cdf((z - half_step) / sigma))


def discrete_var(A,
                 C,
                 grid_sizes=None,
                 std_devs=np.sqrt(10),
                 sim_length=1_000_000,
                 rv=None,
                 order='C',
                 random_state=None):
    r"""
    Generate an `MarkovChain` instance that discretizes a multivariate
    autorregressive process by a simulation of the process.

    This function discretizes a VAR(1) process of the form:

    .. math::

        x_t = A x_{t-1} + C u_t

    where :math:`{u_t}` is drawn iid from a distribution with mean 0 and
    unit standard deviaiton; and :math:`{C}` is a volatility matrix.
    Internally, from this process a sample time series of length
    `sim_length` is produced, and with a cartesian grid as specified by
    `grid_sizes` and `std_devs` a Markov chain is estimated that fits
    the time series, where the states that are never visited under the
    simulation are removed.

    For a mathematical derivation check *Finite-State Approximation Of
    VAR Processes: A Simulation Approach* by Stephanie Schmitt-Grohé and
    Martín Uribe, July 11, 2010. In particular, we follow Schmitt-Grohé
    and Uribe's method in contructing the grid for approximation.

    Parameters
    ----------
    A : array_like(float, ndim=2)
        An m x m matrix containing the process' autocorrelation
        parameters. Its eigenvalues are assumed to have moduli 
        bounded by unity.
    C : array_like(float, ndim=2)
        An m x r volatility matrix
    grid_sizes : array_like(int, ndim=1), optional(default=None)
        An m-vector containing the number of grid points in the
        discretization of each dimension of x_t. If None, then set to
        (10, ..., 10).
    std_devs : float, optional(default=np.sqrt(10))
        The number of standard deviations the grid should stretch in
        each dimension, where standard deviations are measured under the
        stationary distribution.
    sim_length : int, optional(default=1_000_000)
        The length of the simulated time series.
    rv : optional(default=None)
        Object that represents the disturbance term u_t. If None, then
        standard normal distribution from numpy.random is used.
        Alternatively, one can pass a "frozen" object of a multivariate
        distribution from `scipy.stats`. It must have a zero mean and
        unit standard deviation (of dimension r).
    order : str, optional(default='C')
        ('C' or 'F') order in which the states in the cartesian grid are
        enumerated.
    random_state : int or np.random.RandomState/Generator, optional
        Random seed (integer) or np.random.RandomState or Generator
        instance to set the initial state of the random number generator
        for reproducibility. If None, a randomly initialized RandomState
        is used.

    Returns
    -------
    mc : MarkovChain
        An instance of the MarkovChain class that stores the transition
        matrix and state values returned by the discretization method,
        in the following attributes:

            `P` : ndarray(float, ndim=2)
                A 2-dim array containing the transition probability
                matrix over the discretized states.

            `state_values` : ndarray(float, ndim=2)
                A 2-dim array containing the state vectors (of dimension
                m) as rows, which are ordered according to the `order`
                option.

    Examples
    --------
    This example discretizes the stochastic process used to calibrate
    the economic model included in "Downward Nominal Wage Rigidity,
    Currency Pegs, and Involuntary Unemployment" by Stephanie
    Schmitt-Grohé and Martín Uribe, Journal of Political Economy 124,
    October 2016, 1466-1514.

    >>> rng = np.random.default_rng(12345)
    >>> A = [[0.7901, -1.3570],
    ...      [-0.0104, 0.8638]]
    >>> Omega = [[0.0012346, -0.0000776],
    ...          [-0.0000776, 0.0000401]]
    >>> C = scipy.linalg.sqrtm(Omega)
    >>> grid_sizes = [21, 11]
    >>> mc = discrete_var(A, C, grid_sizes, random_state=rng)
    >>> mc.P.shape
    (145, 145)
    >>> mc.state_values.shape
    (145, 2)
    >>> mc.state_values[:10]  # First 10 states
    array([[-0.38556417,  0.02155098],
           [-0.38556417,  0.03232648],
           [-0.38556417,  0.04310197],
           [-0.38556417,  0.05387746],
           [-0.34700776,  0.01077549],
           [-0.34700776,  0.02155098],
           [-0.34700776,  0.03232648],
           [-0.34700776,  0.04310197],
           [-0.34700776,  0.05387746],
           [-0.30845134,  0.        ]])
    >>> mc.simulate(10, random_state=rng)
    array([[ 0.11566925, -0.01077549],
           [ 0.11566925, -0.01077549],
           [ 0.15422567,  0.        ],
           [ 0.15422567,  0.        ],
           [ 0.15422567, -0.01077549],
           [ 0.11566925, -0.02155098],
           [ 0.11566925, -0.03232648],
           [ 0.15422567, -0.03232648],
           [ 0.15422567, -0.03232648],
           [ 0.19278209, -0.03232648]])

    The simulation below uses the same parameters with :math:`{u_t}`
    drawn from a multivariate t-distribution

    >>> df = 100
    >>> Sigma = np.diag(np.tile((df-2)/df, 2))
    >>> mc = discrete_var(A, C, grid_sizes, 
    ...              rv=scipy.stats.multivariate_t(shape=Sigma, df=df), 
    ...              random_state=rng)
    >>> mc.P.shape
    (146, 146)
    >>> mc.state_values.shape
    (146, 2)
    >>> mc.state_values[:10]
    array([[-0.38556417,  0.02155098],
           [-0.38556417,  0.03232648],
           [-0.38556417,  0.04310197],
           [-0.38556417,  0.05387746],
           [-0.34700776,  0.01077549],
           [-0.34700776,  0.02155098],
           [-0.34700776,  0.03232648],
           [-0.34700776,  0.04310197],
           [-0.34700776,  0.05387746],
           [-0.30845134,  0.        ]])
    >>> mc.simulate(10, random_state=rng)
    array([[-0.03855642, -0.02155098],
           [ 0.03855642, -0.03232648],
           [ 0.07711283, -0.03232648],
           [ 0.15422567, -0.03232648],
           [ 0.15422567, -0.04310197],
           [ 0.15422567, -0.03232648],
           [ 0.15422567, -0.03232648],
           [ 0.2313385 , -0.04310197],
           [ 0.2313385 , -0.03232648],
           [ 0.26989492, -0.03232648]])
    """
    A = np.asarray(A)
    C = np.asarray(C)
    m, r = C.shape

    # Run simulation to compute transition probabilities
    random_state = check_random_state(random_state)

    if rv is None:
        u = random_state.standard_normal(size=(sim_length-1, r))
    else:
        u = rv.rvs(size=sim_length-1, random_state=random_state)
    
    v = C @ u.T
    x0 = np.zeros(m)
    X = simulate_linear_model(A, x0, v, ts_length=sim_length)

    # Compute stationary variance-covariance matrix of AR process and use
    # it to obtain grid bounds.
    Sigma = solve_discrete_lyapunov(A, C @ C.T)
    sigma_vector = np.sqrt(np.diagonal(Sigma))    # Stationary std dev
    upper_bounds = std_devs * sigma_vector

    # Build the individual grids along each dimension
    if grid_sizes is None:
        # Set the size of every grid to default_grid_size
        default_grid_size = 10
        grid_sizes = np.full(m, default_grid_size)

    V = [np.linspace(-upper_bounds[i], upper_bounds[i], grid_sizes[i])
         for i in range(m)]

    # Fit the Markov chain
    mc = fit_discrete_mc(X.T, V, order=order)

    return mc



"""
Tests for arma.py file.  Most of this testing can be considered
covered by the numpy tests since we rely on much of their code.

"""
import numpy as np
from numpy.testing import assert_array_equal, assert_
from quantecon import ARMA


class TestARMA():
    def setup_method(self):
        # Initial Values
        phi = np.array([.95, -.4, -.4])
        theta = np.zeros(3)
        sigma = .15


        self.lp = ARMA(phi, theta, sigma)

    def teardown_method(self):
        del self.lp

    def test_simulate(self):
        lp = self.lp

        sim = lp.simulation(ts_length=250)

        assert_(sim.size == 250)

    def test_simulate_with_seed(self):
        lp = self.lp
        seed = 5
        sim0 = lp.simulation(ts_length=10, random_state=seed)
        sim1 = lp.simulation(ts_length=10, random_state=seed)

        assert_array_equal(sim0, sim1)

    def test_impulse_response(self):
        lp = self.lp

        imp_resp = lp.impulse_response(impulse_length=75)

        assert_(imp_resp.size == 75)



"""
Tests for lemke_howson.py
"""
import numpy as np
from numpy.testing import assert_allclose, assert_, assert_raises
from quantecon.game_theory import Player, NormalFormGame, lemke_howson


class TestLemkeHowson():
    def setup_method(self):
        self.game_dicts = []

        # From von Stengel 2007 in Algorithmic Game Theory
        bimatrix = [[(3, 3), (3, 2)],
                    [(2, 2), (5, 6)],
                    [(0, 3), (6, 1)]]
        NEs_dict = {0: ([1, 0, 0], [1, 0]),
                    1: ([0, 1/3, 2/3], [1/3, 2/3])}  # init_pivot: NE
        d = {'g': NormalFormGame(bimatrix),
             'NEs_dict': NEs_dict}
        self.game_dicts.append(d)

    def test_lemke_howson(self):
        for d in self.game_dicts:
            for k in d['NEs_dict'].keys():
                NE_computed = lemke_howson(d['g'], init_pivot=k)
                for action_computed, action in zip(NE_computed,
                                                   d['NEs_dict'][k]):
                    assert_allclose(action_computed, action)


class TestLemkeHowsonDegenerate():
    def setup_method(self):
        self.game_dicts = []

        # From von Stengel 2007 in Algorithmic Game Theory
        bimatrix = [[(3, 3), (3, 3)],
                    [(2, 2), (5, 6)],
                    [(0, 3), (6, 1)]]
        NEs_dict = {0: ([0, 1/3, 2/3], [1/3, 2/3])}
        d = {'g': NormalFormGame(bimatrix),
             'NEs_dict': NEs_dict,
             'converged': True}
        self.game_dicts.append(d)

        # == Examples of cycles by "ad hoc" tie breaking rules == #

        # Example where tie breaking that picks the variable with
        # the smallest row index in the tableau leads to cycling
        A = np.array([[0, 0, 0],
                      [0, 1, 1],
                      [1, 1, 0]])
        B = np.array([[1, 0, 1],
                      [1, 1, 0],
                      [0, 0, 2]])
        NEs_dict = {0: ([0, 2/3, 1/3], [0, 1, 0])}
        d = {'g': NormalFormGame((Player(A), Player(B))),
             'NEs_dict': NEs_dict,
             'converged': True}
        self.game_dicts.append(d)

        # Example where tie breaking that picks the variable with
        # the smallest variable index in the tableau leads to cycling
        perm = [2, 0, 1]
        C = A[:, perm]
        D = B[perm, :]
        NEs_dict = {0: ([0, 2/3, 1/3], [0, 0, 1])}
        d = {'g': NormalFormGame((Player(C), Player(D))),
             'NEs_dict': NEs_dict,
             'converged': True}
        self.game_dicts.append(d)

    def test_lemke_howson_degenerate(self):
        for d in self.game_dicts:
            for k in d['NEs_dict'].keys():
                NE_computed, res = lemke_howson(d['g'], init_pivot=k,
                                                full_output=True)
                for action_computed, action in zip(NE_computed,
                                                   d['NEs_dict'][k]):
                    assert_allclose(action_computed, action)
                assert_(res.converged == d['converged'])


def test_lemke_howson_capping():
    bimatrix = [[(3, 3), (3, 2)],
                [(2, 2), (5, 6)],
                [(0, 3), (6, 1)]]
    g = NormalFormGame(bimatrix)
    m, n = g.nums_actions
    max_iter = 10**6  # big number

    for k in range(m+n):
        NE0, res0 = lemke_howson(g, init_pivot=k, max_iter=max_iter,
                                 capping=None, full_output=True)
        NE1, res1 = lemke_howson(g, init_pivot=k, max_iter=max_iter,
                                 capping=max_iter, full_output=True)
        for action0, action1 in zip(NE0, NE1):
            assert_allclose(action0, action1)
        assert_(res0.init == res1.init)

    init_pivot = 1
    max_iter = m+n
    NE, res = lemke_howson(g, init_pivot=init_pivot, max_iter=max_iter,
                           capping=1, full_output=True)
    assert_(res.num_iter == max_iter)
    assert_(res.init == init_pivot-1)


def test_lemke_howson_invalid_g():
    bimatrix = [[(3, 3), (3, 2)],
                [(2, 2), (5, 6)],
                [(0, 3), (6, 1)]]
    assert_raises(TypeError, lemke_howson, bimatrix)


def test_lemke_howson_invalid_init_pivot_integer():
    bimatrix = [[(3, 3), (3, 2)],
                [(2, 2), (5, 6)],
                [(0, 3), (6, 1)]]
    g = NormalFormGame(bimatrix)
    assert_raises(ValueError, lemke_howson, g, -1)


def test_lemke_howson_invalid_init_pivot_float():
    bimatrix = [[(3, 3), (3, 2)],
                [(2, 2), (5, 6)],
                [(0, 3), (6, 1)]]
    g = NormalFormGame(bimatrix)
    assert_raises(TypeError, lemke_howson, g, 1.0)



import numpy as np
from .normal_form_game import Player
from ..util import check_random_state, rng_integers
from .random import random_pure_actions


class BRD:
    """
    Class representing the best response dynamics model.

    Parameters
    ----------
    payoff_matrix : array_like(float, ndim=2)
        The payoff matrix of the symmetric two-player game.

    N : scalar(int)
        The number of players.

    Attributes
    ----------
    N : scalar(int)
        The number of players.

    num_actions : scalar(int)
        The number of actions.

    player : Player
        Player instance in the model.

    """
    def __init__(self, payoff_matrix, N):
        A = np.asarray(payoff_matrix)
        if A.ndim != 2 or A.shape[0] != A.shape[1]:
            raise ValueError('payoff matrix must be square')
        self.num_actions = A.shape[0]  # Number of actions
        self.N = N  # Number of players

        self.player = Player(A)  # "Representative player"
        self.tie_breaking = 'smallest'

    def _set_action_dist(self, actions):
        if self.N != len(actions):
            raise ValueError("length of `actions` must equal to the number of \
                              players")
        action_dist = np.zeros(self.num_actions)
        for i in range(self.N):
            action_dist[actions[i]] += 1
        return action_dist

    def play(self, action, action_dist, **options):
        """
        Return a new action distribution.

        Parameters
        ----------
        action : scalar(int)
            Pure action of player who takes action.

        action_dist : ndarray(int)
            The initial action distribution of players.

        **options : Keyword arguments passed to the best response method.

        Returns
        -------
        ndarray(int)
            New action distribution.

        """
        tie_breaking = options.get('tie_breaking', self.tie_breaking)
        tol = options.get('tol', None)
        random_state = check_random_state(options.get('random_state', None))

        action_dist[action] -= 1
        next_action = self.player.best_response(action_dist,
                                                tol=tol,
                                                tie_breaking=tie_breaking,
                                                random_state=random_state)
        action_dist[next_action] += 1
        return action_dist

    def time_series(self, ts_length, init_action_dist=None, **options):
        """
        Return a time series of action distributions. The order of player who
        takes a action is randomly choosed.

        Parameters
        ----------
        ts_length : scalar(int)
            The length of time series.

        init_action_dist : array_like(int), optional(default=None)
            The initial action distribution. If None, determined randomly.

        **options : Keyword arguments passed to the best response method and
                    other methods.

        Returns
        -------
        ndarray(int)
            The array representing time series of action distributions.

        """
        tie_breaking = options.get('tie_breaking', self.tie_breaking)
        tol = options.get('tol', None)
        random_state = check_random_state(options.get('random_state', None))

        if init_action_dist is None:
            nums_actions = tuple([self.num_actions] * self.N)
            init_actions = random_pure_actions(nums_actions, random_state)
            init_action_dist = self._set_action_dist(init_actions)

        out = np.empty((ts_length, self.num_actions), dtype=int)
        random_state = check_random_state(random_state)
        player_ind_seq = rng_integers(random_state, self.N, size=ts_length)
        action_dist = np.asarray(init_action_dist)
        for t in range(ts_length):
            out[t, :] = action_dist[:]
            action = np.searchsorted(action_dist.cumsum(), player_ind_seq[t],
                                     side='right')
            random_state = check_random_state(random_state)
            action_dist = self.play(action, action_dist,
                                    tol=tol,
                                    tie_breaking=tie_breaking,
                                    random_state=random_state)
        return out


class KMR(BRD):
    """
    Class representing the Kandori-Mailath-Rob model.

    Parameters
    ----------
    payoff_matrix : array_like(float, ndim=2)
        The payoff matrix of the symmetric two-player game.

    N : scalar(int)
        The number of players.

    epsilon : scalar(float), default=0.1
        The probability of strategy flips.

    Attributes
    ----------
    N : scalar(int)
        See parameters.

    num_actions : scalar(int)
        The number of actions.

    player : Player
        Player instance in the model.

    epsilon : scalar(float)
        See parameters.

    """
    def __init__(self, payoff_matrix, N, epsilon=0.1):
        BRD.__init__(self, payoff_matrix, N)

        # Mutation probability
        self.epsilon = epsilon

    def play(self, action, action_dist, **options):
        """
        See `play` in `BRD`.

        """
        tie_breaking = options.get('tie_breaking', self.tie_breaking)
        tol = options.get('tol', None)
        random_state = check_random_state(options.get('random_state', None))

        if random_state.random() < self.epsilon:  # Mutation
            action_dist[action] -= 1
            random_state = check_random_state(random_state)
            next_action = self.player.random_choice(random_state=random_state)
            action_dist[next_action] += 1
        else:  # Best response
            random_state = check_random_state(random_state)
            action_dist = BRD.play(self, action, action_dist,
                                   tol=tol,
                                   tie_breaking=tie_breaking,
                                   random_state=random_state)
        return action_dist


class SamplingBRD(BRD):
    """
    Class representing the sampling BRD model.

    Parameters
    ----------
    payoff_matrix : array_like(float, ndim=2)
        The payoff matrix of the symmetric two-player game.

    N : scalar(int)
        The number of players.

    k : scalar(int), default=2
        Sample size.

    Attributes
    ----------
    N : scalar(int)
        See parameters.

    num_actions : scalar(int)
        The number of actions.

    player : Player
        Player instance in the model.

    k : scalar(int), default=2
        See parameters.

    """
    def __init__(self, payoff_matrix, N, k=2):
        BRD.__init__(self, payoff_matrix, N)

        # Sample size
        self.k = k

    def play(self, action, action_dist, **options):
        """
        See `play` in `BRD`.

        """
        tie_breaking = options.get('tie_breaking', self.tie_breaking)
        tol = options.get('tol', None)
        random_state = check_random_state(options.get('random_state', None))

        action_dist[action] -= 1
        actions = random_state.choice(self.num_actions, size=self.k,
                                      replace=True, p=action_dist/(self.N-1))
        sample_action_dist = np.bincount(actions, minlength=self.num_actions)
        random_state = check_random_state(random_state)
        next_action = self.player.best_response(sample_action_dist,
                                                tol=tol,
                                                tie_breaking=tie_breaking,
                                                random_state=random_state)
        action_dist[next_action] += 1
        return action_dist



"""
Provides a class called LQ for solving linear quadratic control
problems, and a class called LQMarkov for solving Markov jump
linear quadratic control problems.

"""
from textwrap import dedent
import numpy as np
from scipy.linalg import solve
from ._matrix_eqn import solve_discrete_riccati, solve_discrete_riccati_system
from .util import check_random_state
from .markov import MarkovChain


class LQ:
    r"""
    This class is for analyzing linear quadratic optimal control
    problems of either the infinite horizon form

    .. math::

        \min \mathbb{E}
        \Big[ \sum_{t=0}^{\infty} \beta^t r(x_t, u_t) \Big]

    with

    .. math::

         r(x_t, u_t) := x_t' R x_t + u_t' Q u_t + 2 u_t' N x_t

    or the finite horizon form

    .. math::

         \min \mathbb{E}
         \Big[
         \sum_{t=0}^{T-1} \beta^t r(x_t, u_t) + \beta^T x_T' R_f x_T
         \Big]

    Both are minimized subject to the law of motion

    .. math::

         x_{t+1} = A x_t + B u_t + C w_{t+1}

    Here :math:`x` is n x 1, :math:`u` is k x 1, :math:`w` is j x 1 and the
    matrices are conformable for these dimensions.  The sequence :math:`{w_t}`
    is assumed to be white noise, with zero mean and
    :math:`\mathbb{E} [ w_t' w_t ] = I`, the j x j identity.

    If :math:`C` is not supplied as a parameter, the model is assumed to be
    deterministic (and :math:`C` is set to a zero matrix of appropriate
    dimension).

    For this model, the time t value (i.e., cost-to-go) function :math:`V_t`
    takes the form

    .. math::

         x' P_T x + d_T

    and the optimal policy is of the form :math:`u_T = -F_T x_T`. In the
    infinite horizon case, :math:`V, P, d` and :math:`F` are all stationary.

    Parameters
    ----------
    Q : array_like(float)
        Q is the payoff (or cost) matrix that corresponds with the
        control variable u and is k x k. Should be symmetric and
        non-negative definite
    R : array_like(float)
        R is the payoff (or cost) matrix that corresponds with the
        state variable x and is n x n. Should be symmetric and
        non-negative definite
    A : array_like(float)
        A is part of the state transition as described above. It should
        be n x n
    B : array_like(float)
        B is part of the state transition as described above. It should
        be n x k
    C : array_like(float), optional(default=None)
        C is part of the state transition as described above and
        corresponds to the random variable today.  If the model is
        deterministic then C should take default value of None
    N : array_like(float), optional(default=None)
        N is the cross product term in the payoff, as above. It should
        be k x n.
    beta : scalar(float), optional(default=1)
        beta is the discount parameter
    T : scalar(int), optional(default=None)
        T is the number of periods in a finite horizon problem.
    Rf : array_like(float), optional(default=None)
        Rf is the final (in a finite horizon model) payoff(or cost)
        matrix that corresponds with the control variable u and is n x
        n.  Should be symmetric and non-negative definite

    Attributes
    ----------
    Q, R, N, A, B, C, beta, T, Rf : see Parameters
    P : array_like(float)
        P is part of the value function representation of
        :math:`V(x) = x'Px + d`
    d : array_like(float)
        d is part of the value function representation of
        :math:`V(x) = x'Px + d`
    F : array_like(float)
        F is the policy rule that determines the choice of control in
        each period.
    k, n, j : scalar(int)
        The dimensions of the matrices as presented above

    """

    def __init__(self, Q, R, A, B, C=None, N=None, beta=1, T=None, Rf=None):
        # == Make sure all matrices can be treated as 2D arrays == #
        converter = lambda X: np.atleast_2d(np.asarray(X, dtype='float'))
        self.A, self.B, self.Q, self.R, self.N = list(map(converter,
                                                          (A, B, Q, R, N)))
        # == Record dimensions == #
        self.k, self.n = self.Q.shape[0], self.R.shape[0]

        self.beta = beta

        if C is None:
            # == If C not given, then model is deterministic. Set C=0. == #
            self.j = 1
            self.C = np.zeros((self.n, self.j))
        else:
            self.C = converter(C)
            self.j = self.C.shape[1]

        if N is None:
            # == No cross product term in payoff. Set N=0. == #
            self.N = np.zeros((self.k, self.n))

        if T:
            # == Model is finite horizon == #
            self.T = T
            self.Rf = np.asarray(Rf, dtype='float')
            self.P = self.Rf
            self.d = 0
        else:
            self.P = None
            self.d = None
            self.T = None

            if (self.C != 0).any() and beta >= 1:
                raise ValueError('beta must be strictly smaller than 1 if ' +
                    'T = None and C != 0.')

        self.F = None

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        m = """\
        Linear Quadratic control system
          - beta (discount parameter)       : {b}
          - T (time horizon)                : {t}
          - n (number of state variables)   : {n}
          - k (number of control variables) : {k}
          - j (number of shocks)            : {j}
        """
        t = "infinite" if self.T is None else self.T
        return dedent(m.format(b=self.beta, n=self.n, k=self.k, j=self.j,
                               t=t))

    def update_values(self):
        """
        This method is for updating in the finite horizon case.  It
        shifts the current value function

        .. math::

             V_t(x) = x' P_t x + d_t

        and the optimal policy :math:`F_t` one step *back* in time,
        replacing the pair :math:`P_t` and :math:`d_t` with
        :math:`P_{t-1}` and :math:`d_{t-1}`, and :math:`F_t` with
        :math:`F_{t-1}`

        """
        # === Simplify notation === #
        Q, R, A, B, N, C = self.Q, self.R, self.A, self.B, self.N, self.C
        P, d = self.P, self.d
        # == Some useful matrices == #
        S1 = Q + self.beta * np.dot(B.T, np.dot(P, B))
        S2 = self.beta * np.dot(B.T, np.dot(P, A)) + N
        S3 = self.beta * np.dot(A.T, np.dot(P, A))
        # == Compute F as (Q + B'PB)^{-1} (beta B'PA + N) == #
        self.F = solve(S1, S2)
        # === Shift P back in time one step == #
        new_P = R - np.dot(S2.T, self.F) + S3
        # == Recalling that trace(AB) = trace(BA) == #
        new_d = self.beta * (d + np.trace(np.dot(P, np.dot(C, C.T))))
        # == Set new state == #
        self.P, self.d = new_P, new_d

    def stationary_values(self, method='doubling'):
        """
        Computes the matrix :math:`P` and scalar :math:`d` that represent
        the value function

        .. math::

             V(x) = x' P x + d

        in the infinite horizon case.  Also computes the control matrix
        :math:`F` from :math:`u = - Fx`. Computation is via the solution
        algorithm as specified by the `method` option (default to the
        doubling algorithm) (see the documentation in
        `matrix_eqn.solve_discrete_riccati`).

        Parameters
        ----------
        method : str, optional(default='doubling')
            Solution method used in solving the associated Riccati
            equation, str in {'doubling', 'qz'}.

        Returns
        -------
        P : array_like(float)
            P is part of the value function representation of
            :math:`V(x) = x'Px + d`
        F : array_like(float)
            F is the policy rule that determines the choice of control
            in each period.
        d : array_like(float)
            d is part of the value function representation of
            :math:`V(x) = x'Px + d`

        """
        # === simplify notation === #
        Q, R, A, B, N, C = self.Q, self.R, self.A, self.B, self.N, self.C

        # === solve Riccati equation, obtain P === #
        A0, B0 = np.sqrt(self.beta) * A, np.sqrt(self.beta) * B
        P = solve_discrete_riccati(A0, B0, R, Q, N, method=method)

        # == Compute F == #
        S1 = Q + self.beta * np.dot(B.T, np.dot(P, B))
        S2 = self.beta * np.dot(B.T, np.dot(P, A)) + N
        F = solve(S1, S2)

        # == Compute d == #
        if self.beta == 1:
            d = 0
        else:
            d = self.beta * np.trace(np.dot(P, np.dot(C, C.T))) / (1 - self.beta)

        # == Bind states and return values == #
        self.P, self.F, self.d = P, F, d

        return P, F, d

    def compute_sequence(self, x0, ts_length=None, method='doubling',
                         random_state=None):
        """
        Compute and return the optimal state and control sequences
        :math:`x_0, ..., x_T` and :math:`u_0,..., u_T`  under the
        assumption that :math:`{w_t}` is iid and :math:`N(0, 1)`.

        Parameters
        ----------
        x0 : array_like(float)
            The initial state, a vector of length n

        ts_length : scalar(int)
            Length of the simulation -- defaults to T in finite case

        method : str, optional(default='doubling')
            Solution method used in solving the associated Riccati
            equation, str in {'doubling', 'qz'}. Only relevant when the
            `T` attribute is `None` (i.e., the horizon is infinite).

        random_state : int or np.random.RandomState/Generator, optional
            Random seed (integer) or np.random.RandomState or Generator
            instance to set the initial state of the random number
            generator for reproducibility. If None, a randomly
            initialized RandomState is used.

        Returns
        -------
        x_path : array_like(float)
            An n x T+1 matrix, where the t-th column represents :math:`x_t`

        u_path : array_like(float)
            A k x T matrix, where the t-th column represents :math:`u_t`

        w_path : array_like(float)
            A j x T+1 matrix, where the t-th column represent :math:`w_t`

        """

        # === Simplify notation === #
        A, B, C = self.A, self.B, self.C

        # == Preliminaries, finite horizon case == #
        if self.T:
            T = self.T if not ts_length else min(ts_length, self.T)
            self.P, self.d = self.Rf, 0

        # == Preliminaries, infinite horizon case == #
        else:
            T = ts_length if ts_length else 100
            if self.P is None:
                self.stationary_values(method=method)

        # == Set up initial condition and arrays to store paths == #
        random_state = check_random_state(random_state)
        x0 = np.asarray(x0)
        x0 = x0.reshape(self.n, 1)  # Make sure x0 is a column vector
        x_path = np.empty((self.n, T+1))
        u_path = np.empty((self.k, T))
        w_path = random_state.standard_normal((self.j, T+1))
        Cw_path = np.dot(C, w_path)

        # == Compute and record the sequence of policies == #
        policies = []
        for t in range(T):
            if self.T:  # Finite horizon case
                self.update_values()
            policies.append(self.F)

        # == Use policy sequence to generate states and controls == #
        F = policies.pop()
        x_path[:, 0] = x0.flatten()
        u_path[:, 0] = - np.dot(F, x0).flatten()
        for t in range(1, T):
            F = policies.pop()
            Ax, Bu = np.dot(A, x_path[:, t-1]), np.dot(B, u_path[:, t-1])
            x_path[:, t] = Ax + Bu + Cw_path[:, t]
            u_path[:, t] = - np.dot(F, x_path[:, t])
        Ax, Bu = np.dot(A, x_path[:, T-1]), np.dot(B, u_path[:, T-1])
        x_path[:, T] = Ax + Bu + Cw_path[:, T]

        return x_path, u_path, w_path


class LQMarkov:
    r"""
    This class is for analyzing Markov jump linear quadratic optimal
    control problems of the infinite horizon form

    .. math::

        \min \mathbb{E}
        \Big[ \sum_{t=0}^{\infty} \beta^t r(x_t, s_t, u_t) \Big]

    with

    .. math::

         r(x_t, s_t, u_t) :=
            (x_t' R(s_t) x_t + u_t' Q(s_t) u_t + 2 u_t' N(s_t) x_t)

    subject to the law of motion

    .. math::

         x_{t+1} = A(s_t) x_t + B(s_t) u_t + C(s_t) w_{t+1}

    Here :math:`x` is n x 1, :math:`u` is k x 1, :math:`w` is j x 1 and the
    matrices are conformable for these dimensions.  The sequence :math:`{w_t}`
    is assumed to be white noise, with zero mean and
    :math:`\mathbb{E} [ w_t' w_t ] = I`, the j x j identity.

    If :math:`C` is not supplied as a parameter, the model is assumed to be
    deterministic (and :math:`C` is set to a zero matrix of appropriate
    dimension).

    The optimal value function :math:`V(x_t, s_t)` takes the form

    .. math::

         x_t' P(s_t) x_t + d(s_t)

    and the optimal policy is of the form :math:`u_t = -F(s_t) x_t`.

    Parameters
    ----------
    Π : array_like(float, ndim=2)
        The Markov chain transition matrix with dimension m x m.
    Qs : array_like(float)
        Consists of m symmetric and non-negative definite payoff
        matrices Q(s) with dimension k x k that corresponds with
        the control variable u for each Markov state s
    Rs : array_like(float)
        Consists of m symmetric and non-negative definite payoff
        matrices R(s) with dimension n x n that corresponds with
        the state variable x for each Markov state s
    As : array_like(float)
        Consists of m state transition matrices A(s) with dimension
        n x n for each Markov state s
    Bs : array_like(float)
        Consists of m state transition matrices B(s) with dimension
        n x k for each Markov state s
    Cs : array_like(float), optional(default=None)
        Consists of m state transition matrices C(s) with dimension
        n x j for each Markov state s. If the model is deterministic
        then Cs should take default value of None
    Ns : array_like(float), optional(default=None)
        Consists of m cross product term matrices N(s) with dimension
        k x n for each Markov state,
    beta : scalar(float), optional(default=1)
        beta is the discount parameter

    Attributes
    ----------
    Π, Qs, Rs, Ns, As, Bs, Cs, beta : see Parameters
    Ps : array_like(float)
        Ps is part of the value function representation of
        :math:`V(x, s) = x' P(s) x + d(s)`
    ds : array_like(float)
        ds is part of the value function representation of
        :math:`V(x, s) = x' P(s) x + d(s)`
    Fs : array_like(float)
        Fs is the policy rule that determines the choice of control in
        each period at each Markov state
    m : scalar(int)
        The number of Markov states
    k, n, j : scalar(int)
        The dimensions of the matrices as presented above

    """

    def __init__(self, Π, Qs, Rs, As, Bs, Cs=None, Ns=None, beta=1):

        # == Make sure all matrices for each state are 2D arrays == #
        def converter(Xs):
            return np.array([np.atleast_2d(np.asarray(X, dtype='float'))
                             for X in Xs])
        self.As, self.Bs, self.Qs, self.Rs = list(map(converter,
                                                      (As, Bs, Qs, Rs)))

        # == Record number of states == #
        self.m = self.Qs.shape[0]
        # == Record dimensions == #
        self.k, self.n = self.Qs.shape[1], self.Rs.shape[1]

        if Ns is None:
            # == No cross product term in payoff. Set N=0. == #
            Ns = [np.zeros((self.k, self.n)) for i in range(self.m)]

        self.Ns = converter(Ns)

        if Cs is None:
            # == If C not given, then model is deterministic. Set C=0. == #
            self.j = 1
            Cs = [np.zeros((self.n, self.j)) for i in range(self.m)]

        self.Cs = converter(Cs)
        self.j = self.Cs.shape[2]

        self.beta = beta

        self.Π = np.asarray(Π, dtype='float')

        self.Ps = None
        self.ds = None
        self.Fs = None

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        m = """\
        Markov Jump Linear Quadratic control system
          - beta (discount parameter)       : {b}
          - T (time horizon)                : {t}
          - m (number of Markov states)     : {m}
          - n (number of state variables)   : {n}
          - k (number of control variables) : {k}
          - j (number of shocks)            : {j}
        """
        t = "infinite"
        return dedent(m.format(b=self.beta, m=self.m, n=self.n, k=self.k,
                               j=self.j, t=t))

    def stationary_values(self, max_iter=1000):
        """
        Computes the matrix :math:`P(s)` and scalar :math:`d(s)` that
        represent the value function

        .. math::

             V(x, s) = x' P(s) x + d(s)

        in the infinite horizon case.  Also computes the control matrix
        :math:`F` from :math:`u = - F(s) x`.

        Parameters
        ----------
        max_iter : scalar(int), optional(default=1000)
            The maximum number of iterations allowed

        Returns
        -------
        Ps : array_like(float)
            Ps is part of the value function representation of
            :math:`V(x, s) = x' P(s) x + d(s)`
        ds : array_like(float)
            ds is part of the value function representation of
            :math:`V(x, s) = x' P(s) x + d(s)`
        Fs : array_like(float)
            Fs is the policy rule that determines the choice of control in
            each period at each Markov state

        """

        # == Simplify notations == #
        beta, Π = self.beta, self.Π
        m, n, k = self.m, self.n, self.k
        As, Bs, Cs = self.As, self.Bs, self.Cs
        Qs, Rs, Ns = self.Qs, self.Rs, self.Ns

        # == Solve for P(s) by iterating discrete riccati system== #
        Ps = solve_discrete_riccati_system(Π, As, Bs, Cs, Qs, Rs, Ns, beta,
                                           max_iter=max_iter)

        # == calculate F and d == #
        Fs = np.array([np.empty((k, n)) for i in range(m)])
        X = np.empty((m, m))
        sum1, sum2 = np.empty((k, k)), np.empty((k, n))
        for i in range(m):
            # CCi = C_i C_i'
            CCi = Cs[i] @ Cs[i].T
            sum1[:, :] = 0.
            sum2[:, :] = 0.
            for j in range(m):
                # for F
                sum1 += beta * Π[i, j] * Bs[i].T @ Ps[j] @ Bs[i]
                sum2 += beta * Π[i, j] * Bs[i].T @ Ps[j] @ As[i]

                # for d
                X[j, i] = np.trace(Ps[j] @ CCi)

            Fs[i][:, :] = solve(Qs[i] + sum1, sum2 + Ns[i])

        ds = solve(np.eye(m) - beta * Π,
                   np.diag(beta * Π @ X).reshape((m, 1))).flatten()

        self.Ps, self.ds, self.Fs = Ps, ds, Fs

        return Ps, ds, Fs

    def compute_sequence(self, x0, ts_length=None, random_state=None):
        """
        Compute and return the optimal state and control sequences
        :math:`x_0, ..., x_T` and :math:`u_0,..., u_T`  under the
        assumption that :math:`{w_t}` is iid and :math:`N(0, 1)`,
        with Markov states sequence :math:`s_0, ..., s_T`

        Parameters
        ----------
        x0 : array_like(float)
            The initial state, a vector of length n

        ts_length : scalar(int), optional(default=None)
            Length of the simulation. If None, T is set to be 100

        random_state : int or np.random.RandomState/Generator, optional
            Random seed (integer) or np.random.RandomState or Generator
            instance to set the initial state of the random number
            generator for reproducibility. If None, a randomly
            initialized RandomState is used.

        Returns
        -------
        x_path : array_like(float)
            An n x T+1 matrix, where the t-th column represents :math:`x_t`

        u_path : array_like(float)
            A k x T matrix, where the t-th column represents :math:`u_t`

        w_path : array_like(float)
            A j x T+1 matrix, where the t-th column represent :math:`w_t`

        state : array_like(int)
            Array containing the state values :math:`s_t` of the sample path

        """

        # === solve for optimal policies === #
        if self.Ps is None:
            self.stationary_values()

        # === Simplify notation === #
        As, Bs, Cs = self.As, self.Bs, self.Cs
        Fs = self.Fs

        random_state = check_random_state(random_state)
        x0 = np.asarray(x0)
        x0 = x0.reshape(self.n, 1)

        T = ts_length if ts_length else 100

        # == Simulate Markov states == #
        chain = MarkovChain(self.Π)
        state = chain.simulate_indices(ts_length=T+1,
                                       random_state=random_state)

        # == Prepare storage arrays == #
        x_path = np.empty((self.n, T+1))
        u_path = np.empty((self.k, T))
        w_path = random_state.standard_normal((self.j, T+1))
        Cw_path = np.empty((self.n, T+1))
        for i in range(T+1):
            Cw_path[:, i] = Cs[state[i]] @ w_path[:, i]

        # == Use policy sequence to generate states and controls == #
        x_path[:, 0] = x0.flatten()
        u_path[:, 0] = - (Fs[state[0]] @ x0).flatten()
        for t in range(1, T):
            Ax = As[state[t]] @ x_path[:, t-1]
            Bu = Bs[state[t]] @ u_path[:, t-1]
            x_path[:, t] = Ax + Bu + Cw_path[:, t]
            u_path[:, t] = - (Fs[state[t]] @ x_path[:, t])
        Ax = As[state[T]] @ x_path[:, T-1]
        Bu = Bs[state[T]] @ u_path[:, T-1]
        x_path[:, T] = Ax + Bu + Cw_path[:, T]

        return x_path, u_path, w_path, state



# This file is not meant for public use and will be removed v0.8.0.
# Use the `quantecon` namespace for importing the objects
# included below.

import warnings
from . import _matrix_eqn


__all__ = ['solve_discrete_lyapunov', 'solve_discrete_riccati',
           'solve_discrete_riccati_system']


def __dir__():
    return __all__


def __getattr__(name):
    if name not in __all__:
        raise AttributeError(
                "`quantecon.matrix_eqn` is deprecated and has no attribute "
                f"'{name}'."
            )

    warnings.warn(f"Please use `{name}` from the `quantecon` namespace, the"
                  "`quantecon.matrix_eqn` namespace is deprecated. You can use"
                  f" the following instead:\n `from quantecon import {name}`.",
                  category=DeprecationWarning, stacklevel=2)

    return getattr(_matrix_eqn, name)



"""
Contain a Numba-jitted linear complementarity problem (LCP) solver based
on Lemke's algorithm.

"""
from collections import namedtuple
import numpy as np
from numba import jit
from .pivoting import _pivoting, _lex_min_ratio_test
from .linprog_simplex import PivOptions, FEA_TOL, TOL_PIV, TOL_RATIO_DIFF


LCPResult = namedtuple(
    'LCPResult', ['z', 'success', 'status', 'num_iter']
)
LCPResult.__doc__ = \
    'namedtuple containing the result from `lcp_lemke`.'


# Delete useless docstring for fields of namedtuple
for field in LCPResult._fields:
    getattr(LCPResult, field).__doc__ = ''


@jit(nopython=True, cache=True)
def lcp_lemke(M, q, d=None, max_iter=10**6, piv_options=PivOptions(),
              tableau=None, basis=None, z=None):
    """
    Solve the linear complementarity problem::

        z >= 0
        M @ z + q >= 0
        z @ (M @ z + q) = 0

    by Lemke's algorithm (with the lexicographic pivoting rule).

    Parameters
    ----------
    M : ndarray(float, ndim=2)
        ndarray of shape (n, n).

    q : ndarray(float, ndim=1)
        ndarray of shape (n,).

    d : ndarray(float, ndim=1), optional
        Covering vector, ndarray of shape (n,). Must be strictly
        positive. If None, default to vector of ones.

    max_iter : int, optional(default=10**6)
        Maximum number of iteration to perform.

    piv_options : PivOptions, optional
        PivOptions namedtuple to set the following tolerance values:

            fea_tol : float
                Feasibility tolerance (default={FEA_TOL}). (Not used in
                this function.)

            tol_piv : float
                Pivot tolerance (default={TOL_PIV}).

            tol_ratio_diff : float
                Tolerance used in the the lexicographic pivoting
                (default={TOL_RATIO_DIFF}).

    tableau : ndarray(float, ndim=2), optional
        Temporary ndarray of shape (n, 2*n+2) to store the tableau.
        Modified in place.

    basis : ndarray(int, ndim=1), optional
        Temporary ndarray of shape (n,) to store the basic variables.
        Modified in place.

    z : ndarray(float, ndim=1), optional
        Output ndarray of shape (n,) to store the solution.

    Returns
    -------
    res : LCPResult
        namedtuple consisting of the fields:

            z : ndarray(float, ndim=1)
                ndarray of shape (n,) containing the solution.

            success : bool
                True if the algorithm succeeded in finding a solution.

            status : int
                An integer representing the exit status of the result::

                    0 : Solution found successfully
                    1 : Iteration limit reached
                    2 : Secondary ray termination

            num_iter : int
                The number of iterations performed.

    Examples
    --------
    >>> M = np.array([[1, 0, 0], [2, 1, 0], [2, 2, 1]])
    >>> q = np.array([-8, -12, -14])
    >>> res = lcp_lemke(M, q)
    >>> res.success
    True
    >>> res.z
    array([8., 0., 0.])
    >>> w = M @ res.z + q
    >>> w
    array([0., 4., 2.])
    >>> res.z @ w
    0.0

    References
    ----------
    * K. G. Murty, Linear Complementarity, Linear and Nonlinear
      Programming, 1988.

    """
    n = M.shape[0]

    success = False
    status = 1
    num_iter = 0

    if z is None:
        z = np.empty(n)

    if (q >= 0).all():  # Trivial case
        z[:] = 0
        success = True
        status = 0
        return LCPResult(z, success, status, num_iter)

    if d is None:
        d = np.ones(n)
    if tableau is None:
        tableau = np.empty((n, 2*n+2))
    if basis is None:
        basis = np.empty(n, dtype=np.int_)

    _initialize_tableau(M, q, d, tableau, basis)

    art_var = 2*n  # Artificial variable
    pivcol = art_var

    # Equivalent to lex_min_ratio_test
    pivrow = 0
    ratio_min = q[0] / d[0]
    for i in range(1, n):
        ratio = q[i] / d[i]
        if ratio <= ratio_min + piv_options.tol_ratio_diff:
            pivrow = i
            ratio = ratio_min

    _pivoting(tableau, pivcol, pivrow)
    basis[pivrow], pivcol = pivcol, pivrow + n
    num_iter += 1

    # Array to store row indices in lex_min_ratio_test
    argmins = np.empty(n, dtype=np.int_)

    while num_iter < max_iter:
        pivrow_found, pivrow = _lex_min_ratio_test(
            tableau, pivcol, 0, argmins,
            piv_options.tol_piv, piv_options.tol_ratio_diff
        )

        if not pivrow_found:  # Ray termination
            success = False
            status = 2
            break

        _pivoting(tableau, pivcol, pivrow)
        basis[pivrow], leaving_var = pivcol, basis[pivrow]
        num_iter += 1

        if leaving_var == art_var:  # Solution found
            success = True
            status = 0
            break
        elif leaving_var < n:
            pivcol = leaving_var + n
        else:
            pivcol = leaving_var - n

    _get_solution(tableau, basis, z)

    return LCPResult(z, success, status, num_iter)


lcp_lemke.__doc__ = lcp_lemke.__doc__.format(
    FEA_TOL=FEA_TOL, TOL_PIV=TOL_PIV, TOL_RATIO_DIFF=TOL_RATIO_DIFF
)


@jit(nopython=True, cache=True)
def _initialize_tableau(M, q, d, tableau, basis):
    """
    Initialize the `tableau` and `basis` arrays in place.

    With covering vector d and artificial variable z0, the LCP is
    written as

        q = w - M z - d z0

    where the variables are ordered as (w, z, z0). Thus,
    `tableaus[:, :n]` stores I, `tableaus[:, n:2*n]` stores -M,
    `tableaus[:, 2*n]` stores -d, and `tableaus[:, -1]` stores q, while
    `basis` stores 0, ..., n-1 (variables w).

    Parameters
    ----------
    M : ndarray(float, ndim=2)
        ndarray of shape (n, n).

    q : ndarray(float, ndim=1)
        ndarray of shape (n,).

    d : ndarray(float, ndim=1)
        ndarray of shape (n,).

    tableau : ndarray(float, ndim=2)
        Empty ndarray of shape (n, 2*n+2) to store the tableau.
        Modified in place.

    basis : ndarray(int, ndim=1)
        Empty ndarray of shape (n,) to store the basic variables.
        Modified in place.

    Returns
    -------
    tableau : ndarray(float, ndim=2)
        View to `tableau`.

    basis : ndarray(int, ndim=1)
        View to `basis`.

    """
    n = M.shape[0]

    tableau[:n, :n] = 0
    for i in range(n):
        tableau[i, i] = 1

    for i in range(n):
        for j in range(n):
            tableau[i, n+j] = -M[i, j]

    for i in range(n):
        tableau[i, 2*n] = -d[i]

    for i in range(n):
        tableau[i, -1] = q[i]

    for i in range(n):
        basis[i] = i

    return tableau, basis


@jit(nopython=True, cache=True)
def _get_solution(tableau, basis, z):
    """
    Fetch the solution from `tableau` and `basis`.

    Parameters
    ----------
    tableau : ndarray(float, ndim=2)
        ndarray of shape (n, 2*n+2) containing the terminal tableau.

    basis : ndarray(int, ndim=1)
        ndarray of shape (n,) containing the terminal basis.

    z : ndarray(float, ndim=1)
        Empty ndarray of shape (n,) to store the solution. Modified in
        place.

    Returns
    -------
    z : ndarray(float, ndim=1)
        View to `z`.

    """
    n = z.size

    z[:] = 0
    for i in range(n):
        if n <= basis[i] < 2*n:
            z[basis[i]-n] = tableau[i, -1]

    return z



"""
Tests for markov/utilities.py

"""
from numpy.testing import assert_array_equal
from quantecon.markov import sa_indices


def test_sa_indices():
    num_states, num_actions = 3, 4
    s_expected = [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2]
    a_expected = [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3]

    s, a = sa_indices(num_states, num_actions)

    for indices, indices_expected in zip([s, a], [s_expected, a_expected]):
        assert_array_equal(indices, indices_expected)
